import numpy as np
from astropy.coordinates import SkyCoord, SkyOffsetFrame, SphericalCosLatDifferential, ICRS
from astropy.time import Time
from astropy.table import vstack
import astropy.units as u
import astropy.constants as const
from astroquery.jplhorizons import Horizons
from .config import test_attr
import os
import spiceypy as spice
import urllib.request
import warnings

def read_obj_data():
    """ Reads online table with physical parameter for selected objects
    at http://devel2.linea.gov.br/~altair.gomes/radius.txt

    Return:
        python dictionary
    """
    obj = {}
    try:
        data = urllib.request.urlopen('http://devel2.linea.gov.br/~altair.gomes/radius.txt')
        for line in data:
            arr = line.split()
            obj[arr[0].decode().lower()] = [float(i) for i in arr[1:]]
    except:
        warnings.warn('Online object data table could not be found. Please check internet connection.')
    return obj


def apparent_mag(H, G, dist, sundist, phase=0.0):
    """ Calculates the Apparent Magnitude

    Parameters:
        H (int, float): Absolute Magnitude
        G (int, float): Slope parameter
        dist (int, float): Observer-Object distance, in AU
        sundist (int, float): Sun-Object distance, in AU
        phase (int, float): Phase Angle: Sun-Target-Observer, in deg

    Return:
        ap_mag (float): Apparent Magnitude
    """
    phi0 = np.exp(-3.33*(np.tan(0.5*phase*u.deg)**0.63))
    phi1 = np.exp(-1.87*(np.tan(0.5*phase*u.deg)**1.22))
    Ha = H - 2.5*np.log10((1.0-G)*phi0 + G*phi1)
    ap_mag = Ha + 5*np.log10(sundist*dist)
    return ap_mag.value


def ephem_kernel(time, target, observer, kernels):
    """Calculate the ephemeris from kernel files

    Parameters:
        time (str, Time): instant to calculate ephemeris
        target (str): IAU (kernel) code of the target
        observer (str): IAU (kernel) code of the observer
        kernels (list, str): list of paths for all the kernels

    Return:
        coord (SkyCoord): ICRS coordinate of the target.
    """
    for kern in kernels:
        spice.furnsh(kern)
    time = Time(time)
    t0 = Time('J2000', scale='tdb')
    if time.isscalar:
        time = Time([time])
    dt = (time - t0)
    delt = 0*u.s
    position1 = np.array(spice.spkpos('0', dt.sec, 'J2000', 'NONE', observer)[0])
    while True:
        ### calculates new time
        tempo = dt - delt
        ### calculates vector Solar System Baricenter -> Object
        position2 = spice.spkpos(target, tempo.sec, 'J2000', 'NONE', '0')[0]
        position = (position1 + position2).T
        ### calculates linear distance Earth Topocenter -> Object
        dist = np.linalg.norm(position, axis=0)*u.km
        ### calculates new light time
        delt = (dist/const.c).decompose()
        ### if difference between new and previous light time is smaller than 0.001 sec, than continue.
        if all(np.absolute(((dt - tempo) - delt).sec) < 0.001):
            break
    coord = SkyCoord(position[0], position[1], position[2], frame='icrs', unit=u.km, representation_type='cartesian', obstime=time)
    spice.kclear()
    coord_rd = SkyCoord(ra=coord.spherical.lon, dec=coord.spherical.lat, distance=coord.spherical.distance, obstime=time)
    if len(coord) == 1:
        return coord_rd[0]
    return coord_rd


class EphemPlanete():
    """ EphemPlanete simulates ephem_planete and fit_d2_ksi_eta.

    Parameters:
        name (str): name of the object for search in the JPL database
        ephem (str):Input file with JD, RA, DEC, distance
        radius (int,float): Object radius, in km (Default: Online database)
        error_ra (int,float): Ephemeris RA*cosDEC error, in mas (Default: Online database)
        error_dec (int,float): Ephemeris DEC error, in mas (Default: Online database)
        mass (int,float): Object Mass, in kg (Default: 0.0)
        H (int,float): Object Absolute Magnitude (Default: NaN)
        G (int,float): Object Phase slope (Default: NaN)
    """
    def __init__(self, name, ephem, **kwargs):
        data = np.loadtxt(ephem, unpack=True)
        self.name = name
        self.time = Time(data[0], format='jd')
        self.ephem = SkyCoord(data[1]*u.deg, data[2]*u.deg, data[3]*u.AU)
        self.__reftime = self.time[0]
        self.min_time = Time(data[0].min(), format='jd')
        self.max_time = Time(data[0].max(), format='jd')
        try:
            data = read_obj_data()
        except:
            data = {}
        radius, error_ra, error_dec = data.get(name.lower(), [0,0,0])
        self.radius = kwargs.get('radius', radius)*u.km
        self.error_ra = kwargs.get('error_ra', error_ra)*u.arcsec
        self.error_dec = kwargs.get('error_dec', error_dec)*u.arcsec
        self.mass = kwargs.get('mass', 0.0)*u.kg
        self.H = kwargs.get('H', np.nan)
        self.G = kwargs.get('G', np.nan)

    def fit_d2_ksi_eta(self, star, log=True):
        """ Fits the on-sky ephemeris position relative to a star
        
        Parameters:
            star (str, SkyCoord):The coordinate of the star in the same frame as the ephemeris.
            log (bool): if fit log is to printed. Default: True
        """
        if type(star) == str:
            star = SkyCoord(star, unit=(u.hourangle, u.deg))
        if hasattr(self, 'star') and self.star.to_string('hmsdms', precision=5) == star.to_string('hmsdms', precision=5):
            return
        self.star = star
        target = self.ephem.transform_to(SkyOffsetFrame(origin=star))  
        da = target.cartesian.y
        dd = target.cartesian.z
        dt = (self.time-self.__reftime)/(self.max_time-self.min_time)

        self.ksi = np.polyfit(dt, da.to(u.km).value, 2)
        self.eta = np.polyfit(dt, dd.to(u.km).value, 2)
        ksi = np.poly1d(self.ksi)
        eta = np.poly1d(self.eta)
        dksi = da.to(u.km).value - ksi(dt)
        deta = dd.to(u.km).value - eta(dt)
        rmsk = np.sqrt(np.mean(np.square(dksi)))
        rmse = np.sqrt(np.mean(np.square(deta)))
        if log:
            print('Fitting ephemeris position relative to star coordinate {}'.format(self.star.to_string('hmsdms')))
            print('ksi = aksi*t\u00b2 + bksi*t + cksi')
            print('eta = aeta*t\u00b2 + beta*t + ceta')
            print('t=(jd-{})/({}-{})'.format(self.__reftime.jd, self.max_time.jd,self.min_time.jd))
            print('        aksi={}'.format(self.ksi[0]))
            print('        bksi={}'.format(self.ksi[1]))
            print('        cksi={}'.format(self.ksi[2]))
            print('        aeta={}'.format(self.eta[0]))
            print('        beta={}'.format(self.eta[1]))
            print('        ceta={}'.format(self.eta[2]))
            print('Residual RMS: ksi={:.3f} km, eta={:.3f} km'.format(rmsk, rmse))

    def get_ksi_eta(self, time, star=None):
        """ Returns the on-sky position of the ephemeris relative to a star.

        Parameters:
            time (int, float):Time from which to calculate the position.
            star (str, SkyCoord):The coordinate of the star in the same frame as the ephemeris.

        Returns:
            ksi, eta (float): on-sky position of the ephemeris relative to a star
        """
        if star:
            self.fit_d2_ksi_eta(star)
        time = Time(time)
        if not time.isscalar:
            if any(time < self.min_time) or any(time > self.max_time):
                raise ValueError('time must be in the interval [{},{}]'.format(self.min_time, self.max_time))
        elif time.isscalar:
            if time < self.min_time or time > self.max_time:
                raise ValueError('time must be in the interval [{},{}]'.format(self.min_time, self.max_time))
        if hasattr(self, 'ksi') and hasattr(self, 'eta'):
            ksi = np.poly1d(self.ksi)
            eta = np.poly1d(self.eta)
            k = ksi((time-self.__reftime)/(self.max_time-self.min_time))
            e = eta((time-self.__reftime)/(self.max_time-self.min_time))
            if hasattr(self, 'offset'):
                dist = self.ephem[0].distance.to(u.km).value
                dksi = np.sin(self.offset.d_lon_coslat)*dist
                deta = np.sin(self.offset.d_lat)*dist
                k = k + dksi
                e = e + deta
            return k.value, e.value
        else:
            raise ValueError('A "star" parameter is missing. Please run fit_d2_ksi_eta first.')

    def apparent_magnitude(self,time):
        """ Calculates the Apparent Diameter

        Parameters:
            time (int, float):Time from which to calculate the magnitude.

        Returns:
            ap_mag (float): Apparent magnitude
        """
        time = Time(time)
        obj = Horizons(id=self.name, id_type='majorbody', location='geo', epochs=time.jd)
        eph = obj.ephemerides(extra_precision=True)
        if 'H' in eph.keys():
            self.H = eph['H'][0]
            self.G = eph['G'][0]
        if len(eph['V']) == 1:
            return eph['V'][0]
        else:
            return eph['V'].tolist()

    def add_offset(self, da_cosdec, ddec):
        """Add an offset to the Ephemeris

        Parameters:
            da_cosdec (int, float):Delta_alpha_cos_delta in mas
            ddec (int, float):Delta_delta in mas
        """
        dadc = test_attr(da_cosdec, float, 'da_cosdec')
        dd = test_attr(ddec, float, 'ddec')
        self.offset = SphericalCosLatDifferential(dadc*u.mas, dd*u.mas, 0.0*u.km)

    def __str__(self):
        """ String representation of the EphemPlanete Class.
        """
        out = 'Ephemeris of {}.\n'.format(self.name)
        out += 'Radius: {:.1f}\n'.format(self.radius)
        out += 'Mass: {:.2e}\n'.format(self.mass)
        out += '\nEphem Planete:\n'
        out += 'Valid from {} until {}\n'.format(self.min_time.iso, self.max_time.iso)
        out += 'Ephem Error: RA*cosDEC: {:.3f}; DEC: {:.3f}\n'.format(self.error_ra, self.error_dec)
        if hasattr(self,'star'):
            out += 'Fitted ephemeris position relative to star coordinate {}\n'.format(self.star.to_string('hmsdms'))
            out += '    ksi = aksi*t\u00b2 + bksi*t + cksi\n'
            out += '    eta = aeta*t\u00b2 + beta*t + ceta\n'
            out += '    t=(jd-{})/({}-{})\n'.format(self.__reftime.jd, self.max_time.jd,self.min_time.jd)
            out += '        aksi={}\n'.format(self.ksi[0])
            out += '        bksi={}\n'.format(self.ksi[1])
            out += '        cksi={}\n'.format(self.ksi[2])
            out += '        aeta={}\n'.format(self.eta[0])
            out += '        beta={}\n'.format(self.eta[1])
            out += '        ceta={}\n'.format(self.eta[2])
            target = self.ephem.transform_to(SkyOffsetFrame(origin=self.star))
            da = -target.cartesian.y
            dd = -target.cartesian.z
            dt = (self.time-self.__reftime)/(self.max_time-self.min_time)
            ksi = np.poly1d(self.ksi)
            eta = np.poly1d(self.eta)
            dksi = da.to(u.km).value - ksi(dt)
            deta = dd.to(u.km).value - eta(dt)
            rmsk = np.sqrt(np.mean(np.square(dksi)))
            rmse = np.sqrt(np.mean(np.square(deta)))
            out += 'Residual RMS: ksi={:.3f} km, eta={:.3f} km\n'.format(rmsk, rmse)
        if hasattr(self, 'offset'):
            out += '\nOffset applied: RA*cosDEC: {:.4f}; DEC: {:.4f}\n'.format(
                self.offset.d_lon_coslat.to(u.arcsec), self.offset.d_lat.to(u.arcsec))
        return out


class EphemJPL():
    """ EphemJPL obtains the ephemeris from Horizons service.

    Parameters:
        name (str): name of the object for search in the JPL database
        id_type (str): type of object options: 'smallbody', 'majorbody'
        (planets but also anything that is not a small body), 'designation',
        'name', 'asteroid_name', 'comet_name', 'id' (Horizons id number),
        or 'smallbody' (find the closest match under any id_type), default: 'smallbody'
        radius (int,float): Object radius, in km (Default: Online database)
        error_ra (int,float): Ephemeris RA*cosDEC error, in mas (Default: Online database)
        error_dec (int,float): Ephemeris DEC error, in mas (Default: Online database)
        mass (int,float): Object Mass, in kg (Default: 0.0)
        H (int,float): Object Absolute Magnitude (Default: NaN)
        G (int,float): Object Phase slope (Default: NaN)
    """
    def __init__(self, name, id_type='smallbody', **kwargs):
        self.name = name
        self.id_type='majorbody'
        try:
            data = read_obj_data()
        except:
            data = {}
        radius, error_ra, error_dec = data.get(name.lower(), [0,0,0])
        self.radius = kwargs.get('radius', radius)*u.km
        self.error_ra = kwargs.get('error_ra', error_ra)*u.arcsec
        self.error_dec = kwargs.get('error_dec', error_dec)*u.arcsec
        self.mass = kwargs.get('mass', 0.0)*u.kg
        self.H = kwargs.get('H', np.nan)
        self.G = kwargs.get('G', np.nan)

    def get_position(self, time):
        """ Returns the geocentric position of the object.

        Parameters:
        time (int, float):Time from which to calculate the position.

        Returns:
        coord (SkyCoord): Astropy SkyCoord object with the coordinate at given time
        """
        time = Time(time)
        if not time.isscalar:
            s = len(time)
            def calc_ephem(i):
                n = 50*i
                k = n+50
                if k > s:
                    k = s
                obj = Horizons(id=self.name, id_type=self.id_type, location='geo', epochs=time[n:k].jd)
                return obj.ephemerides(extra_precision=True)
            plus = 0
            if s%50 > 0:
                plus = 1
            eph = vstack([calc_ephem(j) for j in range(s//50+1*plus)])
        else:
            obj = Horizons(id=self.name, id_type=self.id_type, location='geo', epochs=time.jd)
            eph = obj.ephemerides(extra_precision=True)
        coord = SkyCoord(eph['RA'], eph['DEC'], eph['delta'], frame='icrs', obstime=time)
        if hasattr(self, 'offset'):
            pos_frame = SkyOffsetFrame(origin=coord)
            new_pos = SkyCoord(lon=self.offset.d_lon_coslat, lat=self.offset.d_lat, distance=coord.distance, frame=pos_frame)
            coord = new_pos.transform_to(ICRS)
        if len(coord) == 1:
            return coord[0]
        return coord

    def apparent_magnitude(self,time):
        """ Calculates the Apparent Diameter

        Parameters:
            time (int, float):Time from which to calculate the magnitude.

        Returns:
            ap_mag (float): Apparent magnitude
        """
        time = Time(time)
        obj = Horizons(id=self.name, id_type='majorbody', location='geo', epochs=time.jd)
        eph = obj.ephemerides(extra_precision=True)
        if 'H' in eph.keys():
            self.H = eph['H'][0]
            self.G = eph['G'][0]
        if len(eph['V']) == 1:
            return eph['V'][0]
        else:
            return eph['V'].tolist()

    def get_ksi_eta(self, time, star):
        """ Returns the on-sky position of the ephemeris relative to a star.

        Parameters:
        time (int, float):Time from which to calculate the position.
        star (str, SkyCoord):The coordinate of the star in the same frame as the ephemeris.

        Returns:
        ksi, eta (float): on-sky position of the ephemeris relative to a star
        """
        time = Time(time)
        if type(star) == str:
            star = SkyCoord(star, unit=(u.hourangle, u.deg))
        coord = self.get_position(time)
        target = coord.transform_to(SkyOffsetFrame(origin=star))
        da = target.cartesian.y
        dd = target.cartesian.z
        return da.to(u.km).value, dd.to(u.km).value

    def  get_pole_position_angle(self,pole,time):
        """ Returns the geocentric position of the object.

        Parameters:
        pole (str, astropy.SkyCoord):Coordinate of the pole ICRS.
        time (int, float): Time from which to calculate the position.

        Returns:
        position_angle (float): Position angle of the pole, in degrees
        aperture_angle (float): Apeture angle of the pole, in degrees
        """
        time = Time(time)
        if type(pole) == str:
            pole = SkyCoord(pole, unit=(u.hourangle, u.deg))
        obj = self.get_position(time)
        position_angle = obj.position_angle(pole).value*u.rad
        aperture_angle = np.arcsin(-1*(np.sin(pole.dec)*np.sin(obj.dec) + np.cos(pole.dec)*np.cos(obj.dec)*np.cos(pole.ra-obj.ra)))
        return position_angle.to('deg'), aperture_angle.to('deg')

    def add_offset(self, da_cosdec, ddec):
        """Add an offset to the Ephemeris

        Parameters:
            da_cosdec (int, float):Delta_alpha_cos_delta in mas
            ddec (int, float):Delta_delta in mas
        """
        dadc = test_attr(da_cosdec, float, 'da_cosdec')
        dd = test_attr(ddec, float, 'ddec')
        self.offset = SphericalCosLatDifferential(dadc*u.mas, dd*u.mas, 0.0*u.km)

    def __str__(self):
        """ String representation of the EphemPlanete Class.
        """
        out = 'Ephemeris of {}.\n'.format(self.name)
        out += 'Radius: {:.1f}\n'.format(self.radius)
        out += 'Mass: {:.2e}\n'.format(self.mass)
        out += '\nEphemeris are downloaded from Horizons website\n'
        if hasattr(self, 'offset'):
            out += 'Offset applied: RA*cosDEC: {:.4f}; DEC: {:.4f}\n'.format(
                self.offset.d_lon_coslat.to(u.arcsec), self.offset.d_lat.to(u.arcsec))
        return out


class EphemKernel():
    """ EphemKernel gets the ephemeris from bsp kernels.

    Parameters:
        name (str): name of the object for search in the JPL database
        code (str): kernel code of the targeting object
        kernels(list): list of paths for kernels
        radius (int,float): Object radius, in km (Default: Online database)
        error_ra (int,float): Ephemeris RA*cosDEC error, in mas (Default: Online database)
        error_dec (int,float): Ephemeris DEC error, in mas (Default: Online database)
        mass (int,float): Object Mass, in kg (Default: 0.0)
        H (int,float): Object Absolute Magnitude (Default: NaN)
        G (int,float): Object Phase slope (Default: NaN)
    """
    def __init__(self, name, code, kernels, **kwargs):
        self.name = name
        self.code = str(code)
        self.meta = {}
        kerns = []
        for arg in kernels:
            spice.furnsh(arg)
            kerns.append(arg.split('/')[-1].split('.')[0].upper())
        spice.kclear()
        self.meta['kernels'] = '/'.join(kerns)
        self.__kernels = kernels
        try:
            data = read_obj_data()
        except:
            data = {}
        radius, error_ra, error_dec = data.get(name.lower(), [0,0,0])
        self.radius = kwargs.get('radius', radius)*u.km
        self.error_ra = kwargs.get('error_ra', error_ra)*u.arcsec
        self.error_dec = kwargs.get('error_dec', error_dec)*u.arcsec
        self.mass = kwargs.get('mass', 0.0)*u.kg
        self.H = kwargs.get('H', np.nan)
        self.G = kwargs.get('G', np.nan)

    def get_position(self, time):
        """ Returns the geocentric position of the object.

        Parameters:
        time (int, float):Time from which to calculate the position.

        Returns:
        coord (SkyCoord): Astropy SkyCoord object with the coordinate at given time
        """
        pos = ephem_kernel(time, self.code, '399', self.__kernels)
        if hasattr(self, 'offset'):
            pos_frame = SkyOffsetFrame(origin=pos)
            new_pos = SkyCoord(lon=self.offset.d_lon_coslat, lat=self.offset.d_lat, distance=pos.distance, frame=pos_frame)
            pos = new_pos.transform_to(ICRS)
        return pos

    def apparent_magnitude(self,time):
        """ Calculates the Apparent Diameter

        Parameters:
            time (int, float):Time from which to calculate the magnitude.

        Returns:
            ap_mag (float): Apparent magnitude
        """
        time = Time(time)

        if np.isnan(self.H) or np.isnan(self.G):
            obj = Horizons(id=self.name, id_type='majorbody', location='geo', epochs=time.jd)
            eph = obj.ephemerides(extra_precision=True)
            if 'H' in eph.keys():
                self.H = eph['H'][0]
                self.G = eph['G'][0]
            if len(eph['V']) == 1:
                return eph['V'][0]
            else:
                return eph['V'].tolist()

        else:
            obs_obj = self.get_position(time)
            sun_obj = ephem_kernel(time, self.code, '10', self.__kernels)

            # Calculates the phase angle between the 2-vectors
            unit_vector_1 = -obs_obj.cartesian.xyz / np.linalg.norm(obs_obj.cartesian.xyz)
            unit_vector_2 = -sun_obj.cartesian.xyz / np.linalg.norm(sun_obj.cartesian.xyz)
            dot_product = np.dot(unit_vector_1, unit_vector_2)
            phase = np.arccos(dot_product).to(u.deg).value

            return apparent_mag(self.H, self.G, obs_obj.distance.to(u.AU).value, sun_obj.distance.to(u.AU).value, phase)

    def get_ksi_eta(self, time, star):
        """ Returns the on-sky position of the ephemeris relative to a star.

        Parameters:
        time (int, float):Time from which to calculate the position.
        star (str, SkyCoord):The coordinate of the star in the same frame as the ephemeris.

        Returns:
        ksi, eta (float): on-sky position of the ephemeris relative to a star
        """
        time = Time(time)
        if type(star) == str:
            star = SkyCoord(star, unit=(u.hourangle, u.deg))
        coord = self.get_position(time)
        target = coord.transform_to(SkyOffsetFrame(origin=star))  
        da = target.cartesian.y
        dd = target.cartesian.z
        return da.to(u.km).value, dd.to(u.km).value

    def  get_pole_position_angle(self,pole,time):
        """ Returns the geocentric position of the object.

        Parameters:
        pole (str, astropy.SkyCoord):Coordinate of the pole ICRS.
        time (int, float): Time from which to calculate the position.

        Returns:
        position_angle (float): Position angle of the pole, in degrees
        aperture_angle (float): Apeture angle of the pole, in degrees
        """
        time = Time(time)
        if type(pole) == str:
            pole = SkyCoord(pole, unit=(u.hourangle, u.deg))
        obj = self.get_position(time)
        position_angle = obj.position_angle(pole).value*u.rad
        aperture_angle = np.arcsin(-1*(np.sin(pole.dec)*np.sin(obj.dec) + np.cos(pole.dec)*np.cos(obj.dec)*np.cos(pole.ra-obj.ra)))
        return position_angle.to('deg'), aperture_angle.to('deg')

    def add_offset(self, da_cosdec, ddec):
        """Add an offset to the Ephemeris

        Parameters:
            da_cosdec (int, float):Delta_alpha_cos_delta in mas
            ddec (int, float):Delta_delta in mas
        """
        dadc = test_attr(da_cosdec, float, 'da_cosdec')
        dd = test_attr(ddec, float, 'ddec')
        self.offset = SphericalCosLatDifferential(dadc*u.mas, dd*u.mas, 0.0*u.km)

    def __str__(self):
        """ String representation of the EphemPlanete Class.
        """
        out = 'Ephemeris of {}.\n'.format(self.name)
        out += 'Radius: {:.1f}\n'.format(self.radius)
        out += 'Mass: {:.2e}\n'.format(self.mass)
        out += '\nEphem Kernel: {} (code={})\n'.format(self.meta['kernels'],self.code)
        out += 'Ephem Error: RA*cosDEC: {:.3f}; DEC: {:.3f}\n'.format(self.error_ra, self.error_dec)
        if hasattr(self, 'offset'):
            out += 'Offset applied: RA*cosDEC: {:.4f}; DEC: {:.4f}\n'.format(
                self.offset.d_lon_coslat.to(u.arcsec), self.offset.d_lat.to(u.arcsec))
        return out



import matplotlib.pyplot as plt
import numpy as np
import astropy.units as u

def draw_ellipse(equatorial_radius, oblateness=0.0, center_f=0.0, center_g=0.0, position_angle=0.0, ax=None, *args, **kwargs):
    """ Plots an ellipse given input parameters

    Parameters:
        radius (float, int): Semi-major axis of the ellipse.
        oblateness (float, int): Oblateness of the ellipse. Default=0.0
        center_x (float, int): Coordinate of the ellipse (abscissa). Default=0.0
        center_y (float, int): Coordinate of the ellipse (ordinate). Default=0.0
        pos_angle (float, int): Pole position angle. Default=0.0
        **kwargs: all other parameters will be parsed directly to matplotlib
    """
    equatorial_radius = np.array(equatorial_radius, ndmin=1)
    oblateness = np.array(oblateness, ndmin=1)
    center_f = np.array(center_f, ndmin=1)
    center_g = np.array(center_g, ndmin=1)
    position_angle = np.array(position_angle, ndmin=1)
    
    theta= np.linspace(-np.pi, np.pi, 1800)
    size_vals, size_theta = np.indices((len(equatorial_radius), len(theta)))

    ax = ax or plt.gca()

    if len(equatorial_radius) == 1:
        if 'color' not in kwargs:
            kwargs['color'] = 'black'
        if 'lw' not in kwargs:
            kwargs['lw'] = 2
    else:
        if 'color' not in kwargs:
            kwargs['color'] = 'gray'
        if 'lw' not in kwargs:
            kwargs['lw'] = 0.1
        if 'alpha' not in kwargs:
            kwargs['alpha'] = 0.1
        if 'zorder' not in kwargs:
            kwargs['zorder'] = 0.5
    for i in np.arange(len(equatorial_radius)):
        circle_x =  equatorial_radius[i]*np.cos(theta)
        circle_y =  equatorial_radius[i]*(1.0-oblateness[i])*np.sin(theta)
        ax.plot(circle_x*np.cos(position_angle[i]*u.deg) + circle_y*np.sin(position_angle[i]*u.deg) + center_f[i],
                 -circle_x*np.sin(position_angle[i]*u.deg) + circle_y*np.cos(position_angle[i]*u.deg) + center_g[i],
                 *args, **kwargs)
    plt.axis('equal')


class ChiSquare():
    """ ChiSquare stores the arrays for all inputs and given chi-square.

    Parameters:
        chi2 (array): Array with all the chi-square values
        **kwargs: any other given input must be an array with the same size as chi2.
            the keyword name will be associated as the variable name of the given data
            
    Example:
    
    chisquare = ChiSquare(chi2, immersion=t1, emersion=t2)
    t1 and t2 must be an array with the same size as chi2.
    the data can be accessed as:
    chisquare.data['immersion']

    """
    def __init__(self,chi2,npts,**kwargs):
        self.__names = ['chi2']
        self.data = {'chi2': chi2}
        data_size = len(chi2)
        self.npts = npts
        nparam = 0
        for item in kwargs.keys():
            if (kwargs[item].var() != 0):
                nparam += 1
            if len(kwargs[item]) != data_size:
                raise ValueError('{} size must have the same size as given chi2')
            self.__names.append(item)
            self.data[item] = kwargs[item]
        self.nparam = nparam

    def get_nsigma(self,sigma=1, key=None):
        """ Determines the interval of the chi-square within the n-th sigma

        Parameters:
            sigma (float, int): Value of sigma to calculate.
            key (str): keyword the user desire to obtain results.

        Return:
            - if a key is given, it returns two values: the mean value within the n-sigma
            and the error bar within the n-sigma.
            - if no key is given, it returns a dictionary with the minimum chi-square,
            the sigma required, the number of points where chi2 < chi2_min + sigma^2,
            and the mean values and errors for all keys.
        """
        values = np.where(self.data['chi2'] < self.data['chi2'].min() + sigma**2)[0]
        output = {'chi2_min': self.data['chi2'].min()}
        output['sigma'] = sigma
        output['n_points'] = len(values)
        for name in self.__names[1:]:
            vmax = self.data[name][values].max()
            vmin = self.data[name][values].min()
            output[name] = [(vmax+vmin)/2.0, (vmax-vmin)/2.0]
        if key is not None:
            if key not in self.__names[1:]:
                raise ValueError('{} is not one of the available keys. Please choose one of {}'
                                 .format(key, self.__names[1:]))
            return output[key]
        return output

    def plot_chi2(self, key=None, ax=None):
        """ Plots an ellipse given input parameters

        Parameters:
            key (str): Key to plot chi square.
            if no key  is given, it will plot for all the keywords.
        """
        sigma_1 = self.get_nsigma(sigma=1)
        sigma_3 = self.get_nsigma(sigma=3)
        if key is not None and (key not in self.__names[1:]):
            raise ValueError('{} is not one of the available keys. Please choose one of {}'
                                 .format(key, self.data.keys()[1:]))
        k = np.where(self.data['chi2'] < sigma_1['chi2_min']+11)
        for name in self.__names[1:]:
            if (key is not None) and (key != name):
                continue
            if key is None or not ax:
                ax = plt.gca()
            ax.plot(self.data[name][k], self.data['chi2'][k], 'k.')
            ax.set_ylim(sigma_1['chi2_min']-1, sigma_1['chi2_min']+10)
            delta = sigma_3[name][1]
            if delta == 0.0: delta = 1.0
            ax.set_xlim(sigma_3[name][0]-3*delta, sigma_3[name][0]+3*delta)
            ax.axhline(sigma_1['chi2_min']+1,linestyle='--',color='red')
            ax.axhline(sigma_3['chi2_min']+9,linestyle='--',color='red')
            #pl.axvline(t_i[chi2.argmin()],linestyle='--',color='red')
            ax.set_xlabel(name,fontsize=20)
            ax.set_ylabel('Chi2',fontsize=20)
            if key is None:
                plt.show()

    def to_file(self, namefile):
        """ Save the data to a file

        Parameters:
            namefile (str): Filename to save the data
        """
        data = np.vstack(([self.data[i] for i in self.__names]))
        np.savetxt(namefile, data.T, fmt='%11.5f')
        f = open(namefile+'.label', 'w')
        for i, name in enumerate(self.__names):
            f.write('Column {}: {}\n'.format(i+1,name))
        f.close()

    def get_values(self, sigma=0.0, key=None):
        """ Returns all values where the chi-square is within the n-th sigma

        Parameters:
            sigma (float, int): Value of sigma to calculate.
            key (str): keyword the user desire to obtain results.

        Return:
            - if a key is given, it returns list with all the values that are within the n-sigma.
            - if no key is given, it returns a dictionary with the list with all the values
            that are within the n-sigma for all keys.
            - if sigma is zero, it returns the parameters for the minimum chi-square instead of a list.
        """
        values = {}
        if sigma == 0.0:
            k = np.argsort(self.data['chi2'])[0]
        else:
            k = np.where(self.data['chi2'] < self.data['chi2'].min() + sigma**2)[0]
        for name in self.__names[1:]:
            values[name] = self.data[name][k]
        return values

    def __str__(self):
        """ String representation of the ChiSquare Object
        """
        sigma_1 = self.get_nsigma(sigma=1)
        sigma_3 = self.get_nsigma(sigma=3)
        output  = 'Minimum chi-square: {:.3f}\n'.format(sigma_1['chi2_min'])
        output += 'Number of fitted points: {}\n'.format(self.npts)
        output += 'Number of fitted parameters: {}\n'.format(self.nparam)
        output += 'Minimum chi-square per degree of freedom: {:.3f}\n'.format(sigma_1['chi2_min']/(self.npts-self.nparam))
        for name in self.__names[1:]:
            output += '\n{}:\n'.format(name)
            output += '    1-sigma: {:.3f} +/- {:.3f}\n'.format(*sigma_1[name])
            output += '    3-sigma: {:.3f} +/- {:.3f}\n'.format(*sigma_3[name])
        return output



from .star import Star
from .ephem import EphemKernel, EphemJPL, EphemPlanete
import astropy.units as u
import astropy.constants as const
from astropy.coordinates import SkyCoord, EarthLocation, Angle, get_sun
from astropy.coordinates import get_moon, GCRS, ITRS, SkyOffsetFrame
from astropy.time import Time
from astropy.table import Table, Row, Column
from astroquery.vizier import Vizier
import numpy as np
import warnings
import os
import matplotlib
import matplotlib.pyplot as plt

class PredictRow(Row):
    """An Astropy Row object modified for Prediction purposes.
    """
    def plot_occ_map(self,**kwargs):
        """
        Parameters:
            radius: The radius of the shadow. If not given it uses saved value

            nameimg (str): Change the name of the imaged saved.
            resolution (int): Cartopy feature resolution. "1" means a resolution of "10m",
                "2" a resolution of "50m" and "3" a resolution of "100m". Default = 2
            states (bool): True to plot the state division of the countries. The states of
                some countries will only be shown depending on the resolution.
            zoom (int, float): Zooms in or out of the map.
            centermap_geo: Center the map given coordinates in longitude and latitude.
                It must be a list with two numbers. Default=None.
            centermap_delta: Displace the center of the map given displacement
                in X and Y, in km. It must be a list with two numbers. Default=None.
            centerproj: Rotates the Earth to show occultation with the center
                projected at a given longitude and latitude.
            labels: Plots text above and below the map with the occultation parameters.
                Default=True.
            meridians: Plots lines representing the meridians for given interval. Default=30 deg
            parallels: Plots lines representing the parallels for given interval. Default=30 deg
            sites: Plots site positions in map. It must be a python dictionary where the key is
                the name of the site, and the value is a list with longitude, latitude, delta_x,
                delta\_y and color. delta_x and delta_y are displacement, in km, from the point
                of the site in the map and the name. color is the color of the point.
            countries: Plots the names of countries. It must be a python dictionary where the key
                is the name of the country and the value is a list with longitude and latitude
                of the lower left part of the text.
            offset: applies an offset to the ephemeris, calculating new CA and instant of CA.
                It is a pair of delta_RA*cosDEC and delta_DEC.
            mapstyle: Define the color style of the map. 1 is the default black and white scale.
                2 is a colored map.
            error: Ephemeris error in mas. It plots a dashed line representing radius + error.
            lncolor: Changes the color of the lines of the error bar.
            ring: It plots a dashed line representing the location of a ring.
                It is given in km, from the center.
            rncolor: Changes the color of ring lines.
            atm: It plots a dashed line representing the location of an atmosphere.
                It is given in km, from the center.
            rncolor: Changes the color of atm lines.
            heights: It plots a circular dashed line showing the locations where the observer
                would observe the occultation at a given height above the horizons.
                This must be a list.
            hcolor: Changes the color of the height lines.
            mapsize: The size of figure, in cm. It must be a list with two values.
                Default = [46.0, 38.0].
            cpoints: Interval for the small points marking the center of shadow,
                in seconds. Default=60.
            ptcolor: Change the color of the center points.
            alpha: The transparency of the night shade, where 0.0 is full transparency
                and 1.0 is full black. Default = 0.2.
            fmt: The format to save the image. It is parsed directly by matplotlib.pyplot.
                Default = 'png'
            dpi: "Dots per inch". It defines the quality of the image. Default = 100.
            lncolor: Changes the color of the line that represents the limits of the shadow over Earth.
            outcolor: Changes the color of the lines that represents the limits of the shadow outside Earth
            nscale: Arbitrary scale for the size for the name of the site.
            cscale Arbitrary scale for the name of the country.
            sscale Arbitrary scale for the size of point of the site.
            pscale: Arbitrary scale for the size of the points that represent the center of the shadow
            arrow (bool): If true, it plots the arrow with the occultation direction.

            Comment: Only one of centermap_geo and centermap_delta can be given
        """
        radius = kwargs.get('radius', self.meta['radius'])
        if 'radius' in kwargs:
            del kwargs['radius']
        plot_occ_map(self.meta['name'], radius, coord=self['ICRS Star Coord at Epoch'], time=self['Epoch'],
                     ca=float(self['C/A']), pa=float(self['P/A']), vel=float(self['Vel']), dist=float(self['Dist']),
                     mag=float(self['G*']), longi=float(self['long']), **kwargs)

class PredictionTable(Table):
    """An Astropy Table object modified for Prediction purposes.
    """
    Row = PredictRow
    def __init__(self, *args, **kwargs):
        if 'time' in kwargs:
            values = {}
            time = Time(kwargs['time'])
            time.delta_ut1_utc = 0.0
            time.format = 'iso'
            if time.isscalar:
                time = Time([time])
            values['Epoch'] = Column(time)
            def coord_fmt(value):
                return value.to_string('hmsdms',precision=5, sep=' ')
            coord = SkyCoord(kwargs['coord_star'], unit=(u.hourangle,u.deg))
            values['ICRS Star Coord at Epoch'] = Column(coord, format=coord_fmt)
            try:
                coord_geo = SkyCoord(kwargs['coord_obj'])
            except:
                coord_geo = SkyCoord(kwargs['coord_obj'], unit=(u.hourangle,u.deg))
            values['Geocentric Object Position'] = Column(coord_geo, format=coord_fmt)
            values['C/A'] = Column(kwargs['ca'], format='5.3f', unit='arcsec')
            values['P/A'] = Column(kwargs['pa'], format='6.2f', unit='deg')
            values['Vel'] = Column(kwargs['vel'], format='-6.2f', unit='km/s')
            values['Dist'] = Column(kwargs['dist'], format='7.3f', unit='AU')
            del(kwargs['time'], kwargs['coord_star'], kwargs['coord_obj'], kwargs['ca'], kwargs['pa'], kwargs['vel'], kwargs['dist'])
            if 'mag' not in kwargs.keys() and 'mag_20' not in kwargs.keys():
                raise ValueError('User must provide "mag" or "mag_20" parameters')
            if 'mag' in kwargs.keys():
                values['G'] = Column(kwargs['mag'], format='6.3f', unit='mag')
                del kwargs['mag']
            else:
                values['G'] = kwargs['mag_20'] - 2.5*np.log10(np.absolute(values['Vel'])/20.0)
                values['G'].unit = 'mag'
                values['G'].format = '6.3f'
            if 'mag_20' in kwargs.keys():
                values['G*'] = Column(kwargs['mag_20'], format='6.3f', unit='mag')
                del kwargs['mag_20']
            else:
                values['G*'] = values['G'] + 2.5*np.log10(np.absolute(values['Vel'])/20.0)
                values['G*'].unit = 'mag'
                values['G*'].format = '6.3f'
            if 'long' in kwargs.keys():
                values['long'] = Column(kwargs['long'], unit='deg', format='3.0f')
                del kwargs['long']
            else:
                values['long'] = Column((coord.ra - time.sidereal_time('mean', 'greenwich')).wrap_at(360*u.deg), unit='deg', format='3.0f')
            if 'loct' in kwargs.keys():
                values['loct'] = Column(kwargs['loct'], unit='hh:mm')
                del kwargs['loct']
            else:
                longi = (coord.ra - time.sidereal_time('mean', 'greenwich')).wrap_at(360*u.deg)
                ntime = time + longi.hour*u.hour
                values['loct'] = Column(['{}'.format(t.iso[11:16]) for t in ntime], unit='hh:mm')
            moon_pos = get_moon(time)
            values['M-G-T'] = Column(moon_pos.separation(coord), unit='deg', format='3.0f')
            sun_pos = get_sun(time)
            values['S-G-T'] = Column(sun_pos.separation(coord), unit='deg', format='3.0f')
            if 'source' in kwargs.keys():
                values['GAIA-DR2 Source ID'] = Column(kwargs['source'])
                del kwargs['source']
            else:
                values['GAIA-DR2 Source ID'] = Column(np.repeat('', len(time)))
            super().__init__(values, **kwargs)
        else:
            super().__init__(*args, **kwargs)

    def __getitem__(self, item):
        """ The redefinition of __getitem__ allows for selecting prediction based on the ISO date of the event
        """
        if isinstance(item, str) and item not in self.colnames:
            col = self['Epoch']
            arr = list([i for i, c in enumerate(col) if c.iso.startswith(item)])
            if len(arr) is 0:
                raise KeyError('No prediction corresponds to time "{}"'.format(item))
            elif len(arr) > 1:
                return self[arr]
            else:
                return self.Row(self, arr[0])
        return super().__getitem__(item)

    @classmethod
    def from_praia(cls, filename, name, **kwargs):
        """ Create a PredictionTable Table reading from a PRAIA table

        INPUT:
            filename (str): path to the PRAIA table file.
            name (str): Name of the Object of the prediction.
            radius (int,float): Object radius. (not required)
                If not given it's searched in online database.
                If not found online, the defaults is set to zero.

        OUTPUT:
            A PredictionTable
        """
        from .ephem import read_obj_data
        occs = {}
        if os.path.isfile(filename) == False:
            raise IOError('File {} not found'.format(filename))
        try:
            dados = np.loadtxt(filename, skiprows=41, usecols=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
                                                               16, 17, 18, 19, 20, 21, 22, 25, 26, 27, 28, 29),
                dtype={'names': ('dia', 'mes', 'ano', 'hor', 'min', 'sec', 'afh', 'afm', 'afs', 'ded', 'dem', 'des', 'afh2', 'afm2', 'afs2',
                                 'ded2', 'dem2', 'des2', 'ca', 'pa', 'vel', 'delta', 'mR', 'mK', 'long', 'loct', 'ora', 'ode'),
                       'formats': ('S30', 'S30', 'S30','S30', 'S30', 'S30','S20', 'S20', 'S20','S20', 'S20', 'S20', 'S20', 'S20', 'S20',
                                   'S20', 'S20', 'S20', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'S20', 'f4', 'f4')}, ndmin=1)
        except:
            raise IOError('{} is not in PRAIA format or does not have any occultation'.format(filename))
            
        f = open(filename, 'r')
        lines = f.readlines()
        f.close()
        
        max_ca = float(lines[14].split()[-2])*u.arcsec

        ################## reading coordinates #################
        coor = np.char.array(dados['afh'], unicode=True)
        for i in ['afm', 'afs', 'ded', 'dem', 'des']:
            coor = np.core.defchararray.add(coor, ' ')
            coor = np.core.defchararray.add(coor, np.char.array(dados[i], unicode=True))
        coord_star = SkyCoord(coor, frame='icrs', unit=(u.hourangle, u.degree))
        
        coor = np.char.array(dados['afh2'], unicode=True)
        for i in ['afm2', 'afs2', 'ded2', 'dem2', 'des2']:
            coor = np.core.defchararray.add(coor, ' ')
            coor = np.core.defchararray.add(coor, np.char.array(dados[i], unicode=True))
        coord_obj = SkyCoord(coor, frame='icrs', unit=(u.hourangle, u.degree))

        ################### reading time ########################
        tim=np.char.array(dados['ano'], unicode=True)
        len_iso = ['-', '-', ' ', ':',':']
        arr = ['mes', 'dia', 'hor', 'min', 'sec']
        for i in np.arange(len(arr)):
            tim = np.core.defchararray.add(tim, len_iso[i]) 
            tim = np.core.defchararray.add(tim, np.char.array(dados[arr[i]], unicode=True))
        time = Time(np.char.array(tim) + '000')

        ############### defining parameters #############
        data = read_obj_data()
        radius, error_ra, error_dec = data.get(name.lower(), [0,0,0])
        radius = kwargs.get('radius', radius)*u.km
        meta = {'name': name, 'radius': radius, 'max_ca': max_ca, 'ephem': lines[17].split()[-1],
                'error_ra': error_ra*1000, 'error_dec': error_dec*1000}
        return cls(time=time, coord_star=coord_star, coord_obj=coord_obj, ca=dados['ca'], pa=dados['pa'],
                   vel=dados['vel'], mag_20=dados['mR'], dist=dados['delta'], long=dados['long'], loct=dados['loct'], meta=meta)

    def to_praia(self, filename):
        """ Write PredictionTable to PRAIA format.

        INPUT:
            filename(str): name of the file to save table
        """
        from .config import praia_occ_head
        f = open(filename, 'w')
        f.write(praia_occ_head.format(max_ca=self.meta['max_ca'].to(u.arcsec), size=len(self), ephem=self.meta.get('ephem', 'ephem')))
        for time, coord, coord_geo, ca, pa, vel, dist, mag, mag_20, longi, loct, md, sd, source in self.iterrows():
            dmag = mag_20-mag
            f.write("\n {} {} {}  {}  {}   {}   {:5.3f}  {:6.2f} {:-6.2f} {:5.2f} {:4.1f} {:-4.1f} {:-4.1f} {:-4.1f}   {:3.0f}. {}       0.0      0.0 ok g2 0    0    0    0    0".
                    format(time.iso[8:10], time.iso[5:7], time.iso[:4], time.iso[11:21].replace(':', ' '), coord.to_string('hmsdms', precision=4, sep=' '),
                           coord_geo.to_string('hmsdms', precision=4, sep=' '), ca, pa, vel, dist, mag_20, dmag, dmag, dmag, longi, loct)
                   )
        f.close()

    def to_ow(self, ow_des, mode='append'):
        """ Write PredictionTable to OccultWatcher feeder update files.
        Tables will be saved in two files: "tableOccult_update.txt" and "LOG.dat"

        INPUT:
            ow_des (str): Occult Watcher designation for the object.
            mode (str): 'append' to append table to already existing file, default
                        'restart' to overwrite existing file.
        """
        from .config import ow_occ_head
        modes = {'append': 'a+', 'restart': 'w'}
        if mode not in modes.keys():
            raise ValueError('mode param must be "append" or "restart".')

        f = open('tableOccult_update.txt', modes[mode])
        f.write(ow_occ_head.format(name=self.meta['name'], ephem=self.meta.get('ephem', 'ephem'), max_ca=self.meta['max_ca'].to(u.arcsec),
                                   size=len(self), radius=self.meta['radius'], ow_des=ow_des))
        for time, coord, coord_geo, ca, pa, vel, dist, mag, mag_20, longi, loct, md, sd, source in self.iterrows():
            dmag = mag_20-mag
            f.write('{} {} {}  {}   {} {}   {} {}   {:5.3f}  {:6.2f} {:-7.3f} {:7.3f} {:4.1f} {:-4.1f}   {:3.0f}. {}  {:4.0f}  {:4.0f}\n'.
                    format(time.iso[8:10], time.iso[5:7], time.iso[:4], time.iso[11:20].replace(':', ' '), coord.ra.to_string('hour', precision=4, sep=' '),
                           coord.dec.to_string('deg', precision=3, sep=' '), coord_geo.ra.to_string('hour', precision=4, sep=' '),
                           coord_geo.dec.to_string('deg', precision=3, sep=' '), ca, pa, vel, dist, mag_20, dmag, longi, loct, self.meta.get('error_ra', 0),
                           self.meta.get('error_dec', 0)))
        f.write(' '+'-'*148+'\n')
        f.close()

        f = open('LOG.dat', modes[mode])
        t = Time.now()
        for time in self['Epoch']:
            t0 = Time(time.isot.split('T')[0] + ' 00:00:00.0')
            dt = (time-t0).jd*24
            f.write('{} {:5s} {}-{:06.3f}\n'.format(t.isot[:-7], ow_des, time.isot.split('T')[0], dt))
        f.close()

    def plot_occ_map(self, **kwargs):
        """
        Parameters:
            radius: The radius of the shadow. If not given it uses saved value

            nameimg (str): Change the name of the imaged saved.
            resolution (int): Cartopy feature resolution. "1" means a resolution of "10m",
                "2" a resolution of "50m" and "3" a resolution of "100m". Default = 2
            states (bool): True to plot the state division of the countries. The states of
                some countries will only be shown depending on the resolution.
            zoom (int, float): Zooms in or out of the map.
            centermap_geo: Center the map given coordinates in longitude and latitude.
                It must be a list with two numbers. Default=None.
            centermap_delta: Displace the center of the map given displacement
                in X and Y, in km. It must be a list with two numbers. Default=None.
            centerproj: Rotates the Earth to show occultation with the center
                projected at a given longitude and latitude.
            labels: Plots text above and below the map with the occultation parameters.
                Default=True.
            meridians: Plots lines representing the meridians for given interval. Default=30 deg
            parallels: Plots lines representing the parallels for given interval. Default=30 deg
            sites: Plots site positions in map. It must be a python dictionary where the key is
                the name of the site, and the value is a list with longitude, latitude, delta_x,
                delta\_y and color. delta_x and delta_y are displacement, in km, from the point
                of the site in the map and the name. color is the color of the point.
            countries: Plots the names of countries. It must be a python dictionary where the key
                is the name of the country and the value is a list with longitude and latitude
                of the lower left part of the text.
            offset: applies an offset to the ephemeris, calculating new CA and instant of CA.
                It is a pair of delta_RA*cosDEC and delta_DEC.
            mapstyle: Define the color style of the map. 1 is the default black and white scale.
                2 is a colored map.
            error: Ephemeris error in mas. It plots a dashed line representing radius + error.
            lncolor: Changes the color of the lines of the error bar.
            ring: It plots a dashed line representing the location of a ring.
                It is given in km, from the center.
            rncolor: Changes the color of ring lines.
            atm: It plots a dashed line representing the location of an atmosphere.
                It is given in km, from the center.
            rncolor: Changes the color of atm lines.
            heights: It plots a circular dashed line showing the locations where the observer
                would observe the occultation at a given height above the horizons.
                This must be a list.
            hcolor: Changes the color of the height lines.
            mapsize: The size of figure, in cm. It must be a list with two values.
                Default = [46.0, 38.0].
            cpoints: Interval for the small points marking the center of shadow,
                in seconds. Default=60.
            ptcolor: Change the color of the center points.
            alpha: The transparency of the night shade, where 0.0 is full transparency
                and 1.0 is full black. Default = 0.2.
            fmt: The format to save the image. It is parsed directly by matplotlib.pyplot.
                Default = 'png'
            dpi: "Dots per inch". It defines the quality of the image. Default = 100.
            lncolor: Changes the color of the line that represents the limits of the shadow over Earth.
            outcolor: Changes the color of the lines that represents the limits of the shadow outside Earth
            nscale: Arbitrary scale for the size for the name of the site.
            cscale Arbitrary scale for the name of the country.
            sscale Arbitrary scale for the size of point of the site.
            pscale: Arbitrary scale for the size of the points that represent the center of the shadow
            arrow (bool): If true, it plots the arrow with the occultation direction.

            Comment: Only one of centermap_geo and centermap_delta can be given
        """
        for i in range(len(self)):
            self[i].plot_occ_map(**kwargs)


def occ_params(star, ephem, time):
    """ Calculates the parameters of the occultation, as instant, CA, PA.
        
    Parameters:
    star (Star): The coordinate of the star in the same frame as the ephemeris.
    It must be a Star object.
    ephem (Ephem): Ephemeris. It must be an Ephemeris object.
    
    Return:
    instant of CA (Time): Instant of Closest Approach
    CA (arcsec): Distance of Closest Approach
    PA (deg): Position Angle at Closest Approach
    vel (km/s): Velocity of the occultation
    dist (AU): the object distance.
    """
    
    delta_t = 0.05
    
    if type(star) != Star:
        raise ValueError('star must be a Star object')
    if type(ephem) not in [EphemKernel, EphemJPL, EphemPlanete]:
        raise ValueError('ephem must be an Ephemeris object')
        
    time = Time(time)
    tt = time + np.arange(-600, 600, delta_t)*u.s
    coord = star.geocentric(tt[0])
    if type(ephem) == EphemPlanete:
        ephem.fit_d2_ksi_eta(coord, log=False)
    
    if type(ephem) == EphemJPL:
        tt = time + np.arange(-600, 600, 4)*u.s
        ksi, eta = ephem.get_ksi_eta(tt, coord)
        dd = np.sqrt(ksi*ksi+eta*eta)
        min = np.argmin(dd)
        tt = tt[min] + np.arange(-8, 8, 0.05)*u.s
    
    ksi, eta = ephem.get_ksi_eta(tt, coord)
    dd = np.sqrt(ksi*ksi+eta*eta)
    min = np.argmin(dd)
    
    if type(ephem) == EphemPlanete:
        dist = ephem.ephem[int(len(ephem.time)/2)].distance
    else:
        dist = ephem.get_position(time).distance
    
    ca = np.arcsin(dd[min]*u.km/dist).to(u.arcsec)
    
    pa = (np.arctan2(ksi[min],eta[min])*u.rad).to(u.deg)
    if pa < 0*u.deg:
        pa = pa + 360*u.deg
    
    dksi = ksi[min+1]-ksi[min]
    deta = eta[min+1]-eta[min]
    vel = np.sqrt(dksi**2 + deta**2)/delta_t
    vel = vel*np.sign(dksi)*(u.km/u.s)
    
    return tt[min], ca, pa, vel, dist.to(u.AU)


def prediction(ephem, time_beg, time_end, mag_lim=None, step=60, divs=1, sigma=1, log=True):
    """ Predicts stellar occultations

    Parameters:
    ephem (Ephem): Ephemeris. It must be an Ephemeris object.
    time_beg (Time): Initial time for prediction
    time_beg (Time): Final time for prediction
    mag_lim (int,float): Faintest Gmag for search
    step (int, float): step, in seconds, of ephem times for search
    divs (int): number of regions the ephemeris will be splitted
        for better search of occultations
    log (bool): To show what is being done at the moment.

    Return:
    occ_params (Table): PredictionTable with the occultation params for each event
    """
    # generate ephemeris
    if type(ephem) is not EphemKernel:
        raise TypeError('At the moment prediction only works with EphemKernel')
    if log: print("Generating Ephemeris ...")
    time_beg = Time(time_beg)
    time_end = Time(time_end)
    dt = np.arange(0, (time_end-time_beg).sec, step)*u.s
    t = time_beg + dt
    coord = ephem.get_position(t)

    # define catalogue parameters
    kwds = {}
    kwds['columns'] = ['Source', 'RA_ICRS', 'DE_ICRS']
    kwds['row_limit']=10000000
    kwds['timeout']=600
    if mag_lim:
        kwds['column_filters']={"Gmag":"<{}".format(mag_lim)}
    vquery = Vizier(**kwds)

    # determine suitable divisions for star search
    radius = ephem.radius + const.R_earth
    mindist = np.arcsin(radius/coord.distance).max() + sigma*np.max([ephem.error_ra.value,ephem.error_dec.value])*u.arcsec

    divisions = np.array_split(np.arange(len(coord)), divs)

    if log: print('Ephemeris was split in {} parts for better search of stars'.format(len(divisions)))

    # makes predictions for each division
    occs = []
    for i,vals in enumerate(divisions):
        if log: print('\nSearching occultations in part {}/{}'.format(i+1,len(divisions)))
        nt = t[vals]
        ncoord = coord[vals]
        ra = np.mean([ncoord.ra.min().deg,ncoord.ra.max().deg])
        dec = np.mean([ncoord.dec.min().deg,ncoord.dec.max().deg])
        width = ncoord.ra.max() - ncoord.ra.min() + 2*mindist
        height = ncoord.dec.max() - ncoord.dec.min() + 2*mindist
        pos_search = SkyCoord(ra*u.deg, dec*u.deg)

        if log: print('Downloading stars ...')
        catalogue = vquery.query_region(pos_search, width=width, height=height, catalog='I/345/gaia2')
        if log: print('Identifying occultations ...')
        if len(catalogue) == 0:
            continue
        catalogue = catalogue[0]
        stars = SkyCoord(catalogue['RA_ICRS'], catalogue['DE_ICRS'])
        idx, d2d, d3d = stars.match_to_catalog_sky(ncoord)

        dist = np.arcsin(radius/ncoord[idx].distance) + sigma*np.max([ephem.error_ra.value,ephem.error_dec.value])*u.arcsec
        k = np.where(d2d < dist)[0]
        for ev in k:
            star = Star(code=catalogue['Source'][ev], nomad=False, log=False)
            c = star.geocentric(nt[idx][ev])
            pars = [star.code, SkyCoord(c.ra,c.dec), star.mag['G']]
            try:
                pars = np.hstack((pars, occ_params(star, ephem, nt[idx][ev])))
                occs.append(pars)
            except:
                pass

    meta = {'name': ephem.name, 'time_beg': time_beg, 'time_end': time_end, 'maglim': mag_lim, 'max_ca': mindist,
            'radius':ephem.radius.to(u.km).value, 'error_ra': ephem.error_ra.to(u.mas).value, 'error_dec': ephem.error_dec.to(u.mas).value}
    if not occs:
        warnings.warn('No stellar occultation was found')
        return PredictionTable(meta=meta)
    # create astropy table with the params
    occs2 = np.transpose(occs)
    time = Time(occs2[3])
    geocentric = ephem.get_position(time)
    k = np.argsort(time)
    t = PredictionTable(time=time[k], coord_star=occs2[1][k], coord_obj=geocentric[k], ca=[i.value for i in occs2[4][k]],
        pa=[i.value for i in occs2[5][k]], vel=[i.value for i in occs2[6][k]], mag=occs2[2][k],
        dist=[i.value for i in occs2[7][k]], source=occs2[0][k], meta=meta)
    if log:
        print('{} occultations found'.format(len(t)))
    return t


def xy2latlon(x, y, loncen, latcen, time):
    """ Calculates the longitude and latitude given projected positions x and y

    Parameters:
        x (int, float): Projected position in x, in the GCRS (meters)
        y (int, float): Projected position in y, in the GCRS (meters)
        loncen (int, float): Center longitude of projection (deg)
        latcen (int, float): Center latitude of projection (deg)
        time (Time): Time of refered projection

    Return:
        lon, lat (float): Longitude and Latitude whose projection at
                          loncen, lat results in x, y. (deg)
    """
    r = const.R_earth.to(u.m).value
    site_cen = EarthLocation(loncen*u.deg,latcen*u.deg)
    itrs_cen = site_cen.get_itrs(obstime=time)
    gcrs_cen = itrs_cen.transform_to(GCRS(obstime=time))

    z = np.array(y, ndmin=1)
    y = np.array(x, ndmin=1)
    x2 = r*r-y*y-z*z

    a = np.where(x2 >= 0.0)
    x = np.sqrt(x2[a])

    y = y[a]
    z = z[a]
    lon = np.repeat(1e+31, len(x2))
    lat = np.repeat(1e+31, len(x2))
    center_frame = SkyOffsetFrame(origin=gcrs_cen)
    if len(x) > 0:
        n = 0
        if not time.isscalar and len(time) == len(x2):
            time = time[a]
        while True:
            n+=1
            new_pos = SkyCoord(x*u.m, y*u.m, z*u.m, representation_type='cartesian', frame=center_frame[a])
            n_coord = new_pos.transform_to(GCRS(obstime=time))
            n_itrs = n_coord.transform_to(ITRS(obstime=time))
            n_site = n_itrs.earth_location
            n_site = EarthLocation(n_site.lon, n_site.lat, 0)
            itrs_site = n_site.get_itrs(obstime=time)
            gcrs_site = itrs_site.transform_to(GCRS(obstime=time))
            target1 = gcrs_site.transform_to(center_frame[a])
            if n == 4:
                lon[a] = n_site.lon.deg
                lat[a] = n_site.lat.deg
                break
            x = target1.cartesian.x.to(u.m).value
    return lon, lat


def latlon2xy(lon, lat, loncen, latcen):
    """ Calculates the projection of longitude and latitude in the loncen, latcen direction.

    Parameters:
        lon (int, float): Longitude to calculate projection
        lat (int, float): Latitude to calculate projection
        loncen (int, float): Center longitude of projection (deg)
        latcen (int, float): Center latitude of projection (deg)

    Return:
        x, y (float): Projection of lon, lat at loncen, latcen, in the ITRS (meters)
    """
    site_cen = EarthLocation(loncen*u.deg,latcen*u.deg)
    itrs_cen = site_cen.get_itrs()

    lon = np.array(lon, ndmin=1)
    lat = np.array(lat, ndmin=1)
    site = EarthLocation(lon*u.deg,lat*u.deg, height=0*u.m)
    itrs_site = site.get_itrs()

    target = itrs_site.transform_to(SkyOffsetFrame(origin=itrs_cen))
    y = target.cartesian.y.to(u.m).value
    z = target.cartesian.z.to(u.m).value
    k = np.where(target.cartesian.x.to(u.m).value < 0.0)
    y[k] = 1e+31
    z[k] = 1e+31
    return y, z


def plot_occ_map(name, radius, **kwargs):
    """ Plots map of the occultation

    Parameters:
        Required params:
        name (str): Name of the object
        radius (int, float): radius of the object (km)
        coord (str, SkyCoord): Coordinate of the star
            ("hh mm ss.sss dd mm ss.sss" or "hh.hhhhhhhh dd.dddddddd")
        time (str, Time): Instant of Closest Approach (iso or isot format)
        ca (int, float): Closest Approach Distance (arcsec)
        pa (int, float): Position Angle at C/A (deg)
        vel (int, vel): Velocity of the event (km/s)
        dist (int, float): Object distance at C/A (AU)

        Map configuration:
        nameimg (str): Change the name of the imaged saved.
        resolution (int): Cartopy feature resolution. "1" means a resolution of "10m",
            "2" a resolution of "50m" and "3" a resolution of "100m". Default = 2
        states (bool): True to plot the state division of the countries. The states of
            some countries will only be shown depending on the resolution.
        zoom (int, float): Zooms in or out of the map.
        centermap_geo: Center the map given coordinates in longitude and latitude.
            It must be a list with two numbers. Default=None.
        centermap_delta: Displace the center of the map given displacement
            in X and Y, in km. It must be a list with two numbers. Default=None.
        centerproj: Rotates the Earth to show occultation with the center
            projected at a given longitude and latitude.
        labels: Plots text above and below the map with the occultation parameters.
            Default=True.
        meridians: Plots lines representing the meridians for given interval. Default=30 deg
        parallels: Plots lines representing the parallels for given interval. Default=30 deg
        sites: Plots site positions in map. It must be a python dictionary where the key is
            the name of the site, and the value is a list with longitude, latitude, delta_x,
            delta\_y and color. delta_x and delta_y are displacement, in km, from the point
            of the site in the map and the name. color is the color of the point.
        countries: Plots the names of countries. It must be a python dictionary where the key
            is the name of the country and the value is a list with longitude and latitude
            of the lower left part of the text.
        offset: applies an offset to the ephemeris, calculating new CA and instant of CA.
            It is a pair of delta_RA*cosDEC and delta_DEC.
        mapstyle: Define the color style of the map. 1 is the default black and white scale.
            2 is a colored map.
        error: Ephemeris error in mas. It plots a dashed line representing radius + error.
        lncolor: Changes the color of the lines of the error bar.
        ring: It plots a dashed line representing the location of a ring.
            It is given in km, from the center.
        rncolor: Changes the color of ring lines.
        atm: It plots a dashed line representing the location of an atmosphere.
            It is given in km, from the center.
        rncolor: Changes the color of atm lines.
        heights: It plots a circular dashed line showing the locations where the observer
            would observe the occultation at a given height above the horizons.
            This must be a list.
        hcolor: Changes the color of the height lines.
        mapsize: The size of figure, in cm. It must be a list with two values.
            Default = [46.0, 38.0].
        cpoints: Interval for the small points marking the center of shadow,
            in seconds. Default=60.
        ptcolor: Change the color of the center points.
        alpha: The transparency of the night shade, where 0.0 is full transparency
            and 1.0 is full black. Default = 0.2.
        fmt: The format to save the image. It is parsed directly by matplotlib.pyplot.
            Default = 'png'
        dpi: "Dots per inch". It defines the quality of the image. Default = 100.
        lncolor: Changes the color of the line that represents the limits of the shadow over Earth.
        outcolor: Changes the color of the lines that represents the limits of the shadow outside Earth
        nscale: Arbitrary scale for the size for the name of the site.
        cscale Arbitrary scale for the name of the country.
        sscale Arbitrary scale for the size of point of the site.
        pscale: Arbitrary scale for the size of the points that represent the center of the shadow
        arrow (bool): If true, it plots the arrow with the occultation direction.

        Comment: Only one of centermap_geo and centermap_delta can be given
    """
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature

    if not type(name) == str:
        raise TypeError('name keyword must be a string')

    radius = radius*u.km

    occs = {}
    if not all([i in kwargs for i in ['coord', 'time', 'ca', 'pa', 'vel', 'dist']]):
        raise KeyError("It's missing input parameters. Please read the tutorial")

    try:
        occs['stars'] = SkyCoord(kwargs['coord'], frame='icrs', unit=(u.hourangle, u.degree))
    except:
        raise KeyError('"star" keyword is not in the format: "hh mm ss.sss dd mm ss.sss" or "hh.hhhhhhhh dd.dddddddd"')

    try:
        occs['datas'] = Time(kwargs['time'])
    except:
        raise KeyError('"time" keyword is not a iso or isot time format')
    occs['ca'] = kwargs['ca']*u.arcsec
    occs['posa'] = kwargs['pa']*u.deg
    occs['vel'] = kwargs['vel']*(u.km/u.s)
    occs['dist'] = kwargs['dist']*u.AU
    occs['magG'] = kwargs.get('mag', 0.0)
    occs['longi'] = kwargs.get('longi', 0.0)

    mapstyle = kwargs.get('mapstyle', 1)
    if mapstyle not in [1,2]:
        raise ValueError('mapstyle must be 1 or 2]')

    resolution = kwargs.get('resolution', 2)
    if resolution not in [1, 2, 3]:
        raise TypeError('resolution keyword must be one of these: [1, 2, 3] where 1=10m, 2=50m and 3=100m')
    res = ['10m', '50m', '110m']
    resolution = res[resolution-1]

    nameimg = kwargs.get('nameimg', '{}_{}'.format(name, occs['datas'].isot))
    fmt = kwargs.get('fmt', 'png')
    dpi = kwargs.get('dpi', 100)
    step = kwargs.get('step', 1)
    mapsize = kwargs.get('mapsize', [46.0, 38.0])*u.cm
    erro = kwargs.get('error', None)
    ring = kwargs.get('ring', None)
    atm = kwargs.get('atm', None)
    cpoints = kwargs.get('cpoints', 60)
    states = kwargs.get('states', True)
    labels = kwargs.get('labels', True)
    meridians = kwargs.get('meridians', 30)
    parallels = kwargs.get('parallels', 30)
    nscale = kwargs.get('nscale', 1)
    cscale = kwargs.get('cscale', 1)
    sscale = kwargs.get('sscale', 1)
    pscale = kwargs.get('pscale', 1)
    heights = np.array(kwargs.get('heights'), None)
    alpha = kwargs.get('alpha', 0.2)
    centermap_geo = kwargs.get('centermap_geo', None)
    centermap_delta = kwargs.get('centermap_delta', None)
    if 'centermap_geo' in kwargs and 'centermap_delta' in kwargs:
        raise ValueError('User must give "centermap_geo" OR "centermap_delta"')
    zoom = kwargs.get('zoom', 1)
    off_ra, off_de = kwargs.get('offset', [0.0, 0.0])*u.mas
    arrow = kwargs.get('arrow', True)

    sites = {}
    if 'sites' in kwargs.keys():
        if type(kwargs['sites']) == str and os.path.isfile(kwargs['sites']):
            data = np.loadtxt(kwargs['sites'], dtype={'names': ('name', 'lon', 'lat', 'offx', 'offy', 'color'),
                                                        'formats': ('S30', 'f8', 'f8', 'f8', 'f8', 'S30')}, delimiter=',', ndmin=1)
            for i, s in enumerate(data):
                sites[s['name'].strip().decode()] = [s['lon'],s['lat'], s['offx'], s['offy'], s['color'].strip().decode() ]
        elif type(kwargs['sites']) == dict:
            sites = kwargs['sites']
        else:
            raise TypeError('sites keyword must be a file or a dictionary')

    countries = {}
    if 'countries' in kwargs.keys():
        if type(kwargs['countries']) == str and os.path.isfile(kwargs['countries']):
            data = np.loadtxt(kwargs['countries'], dtype={'names': ('name', 'lon', 'lat'), 'formats': ('S30', 'f8', 'f8')}, delimiter=',', ndmin=1)
            for i, c in enumerate(data):
                countries[c['name'].strip().decode()] = [c['lon'],c['lat']]
        elif type(kwargs['countries']) == dict:
            countries = kwargs['countries']
        else:
            raise TypeError('country keyword must be a file or a dictionary')

########## calculates offsets ######
    dca = off_ra*np.sin(occs['posa']) + off_de*np.cos(occs['posa'])
    dt = (-(off_ra*np.cos(occs['posa']) - off_de*np.sin(occs['posa'])).to(u.rad)*occs['dist'].to(u.km)/np.absolute(occs['vel'])).value*u.s
    ca1 = occs['ca'] + dca
    data = occs['datas'] + dt

##### define map parameters #####
    center_gcrs = GCRS(occs['stars'].ra, occs['stars'].dec, 1*u.R_earth, obstime=data)
    center_itrs = center_gcrs.transform_to(ITRS(obstime=data))
    center_map = center_itrs.earth_location
    centert = True
    if 'centerproj' in kwargs.keys():
        if type(kwargs['centerproj']) == EarthLocation:
            center_map = kwargs['centerproj']
        elif np.array(kwargs['centerproj']).shape == (2,):
            center_map = EarthLocation.from_geodetic(*kwargs['centerproj'], 0.0)
        else:
            raise TypeError('centerproj must be an Astropy EarthLocation Object or an array with Longitude and Latitude only')
        centert = False
    fig = plt.figure(figsize=(mapsize.to(u.imperial.inch).value))
    projection = ccrs.Orthographic(central_longitude=center_map.lon.value, central_latitude=center_map.lat.value)
    if labels:
        axf = plt.axes(projection=projection)
    else:
        axf = plt.axes([-0.001,-0.001,1.002,1.002], projection=projection)
    axf.set_global()

### calculates regions for zoom
    limits = None
    r = const.R_earth.to(u.m).value
    if centermap_geo is not None:
        cx, cy = latlon2xy(centermap_geo[0], centermap_geo[1], center_map.lon.value, center_map.lat.value)
        limits = [cx/1000.0, cy/1000.0]
        if np.any(np.absolute(limits) > r):
            raise ValueError('Coordinates for centermap_geo are outside the visible range.')
    elif centermap_delta is not None:
        limits = centermap_delta
    elif zoom is not 1:
        limits = [0,0]
    if limits is not None:
        dr = r/zoom
        l0 = (limits[0]*u.km).to(u.m).value
        l1 = (limits[1]*u.km).to(u.m).value
        dmsize = mapsize[0]/mapsize[1]
        if mapsize[1] < mapsize[0]:
            lx = l0 - dr*dmsize
            ux = l0 + dr*dmsize
            ly = l1 - dr
            uy = l1 + dr
        else:
            lx = l0 - dr
            ux = l0 + dr
            ly = l1 - dr/dmsize
            uy = l1 + dr/dmsize
        axf.set_xlim(lx,ux)
        axf.set_ylim(ly,uy)
        if labels and zoom > 1:
            centert = False

### plots features
    axf.coastlines(resolution=resolution, color='0.3')
    ocean = cfeature.NaturalEarthFeature('physical', 'ocean', resolution)
    land = cfeature.NaturalEarthFeature('physical', 'land', resolution)
    border = cfeature.NaturalEarthFeature('cultural', 'admin_0_countries', resolution)
    if mapstyle == 1:
        axf.add_feature(ocean, zorder=0, color='0.9')
        axf.add_feature(land, zorder=0, edgecolor='None', color='1.0')
        axf.add_feature(border, zorder=0.1, edgecolor='0.4', facecolor='None')
        axf.add_feature(cfeature.RIVERS, zorder=0, edgecolor='0.7')
        axf.add_feature(cfeature.LAKES, zorder=0, color='0.7')
        ptcolor= 'black'
        lncolor= 'blue'
        ercolor= 'blue'
        rncolor= 'blue'
        atcolor= 'blue'
        outcolor= 'red'
        hcolor = 'black'
    elif mapstyle == 2:
        axf.add_feature(ocean, zorder=0, facecolor=cfeature.COLORS['water'])
        axf.add_feature(land, zorder=0, edgecolor='None', facecolor=cfeature.COLORS['land'])
        axf.add_feature(border, zorder=0, edgecolor='0.5', facecolor=cfeature.COLORS['land'])
        axf.add_feature(border, zorder=0.1, edgecolor='0.5', facecolor='None')
        axf.add_feature(cfeature.RIVERS, zorder=0)
        axf.add_feature(cfeature.LAKES, zorder=0)
        ptcolor= 'red'
        lncolor= 'blue'
        ercolor= 'red'
        rncolor= 'black'
        atcolor= 'black'
        outcolor= 'red'
        hcolor = 'black'
    if states:
        states_r = cfeature.NaturalEarthFeature('cultural', 'admin_1_states_provinces', resolution)
        axf.add_feature(states_r, zorder=0, edgecolor='0.6', facecolor='None')

    gl = axf.gridlines(xlocs=np.arange(-180,180.001,meridians), ylocs=np.arange(-90,90.001,parallels))
    gl.n_steps = 180
    sun = get_sun(data)
    sun_lat = sun.dec
    sun_lon = sun.ra - data.sidereal_time('mean', 'greenwich')
    pole_lon = sun_lon.deg
    pole_lat = sun_lat.deg
    proj_sun = ccrs.Orthographic(central_longitude=pole_lon+180,central_latitude=-pole_lat)
    bordx = r*np.cos(np.arange(0,361,0.5)*u.deg)
    bordy = r*np.sin(np.arange(0,361,0.5)*u.deg)
    axf.fill(bordx,bordy, transform=proj_sun, linewidth=0, color='black', alpha=alpha)
    axf.fill(bordx*np.cos(18*u.deg),bordy*np.cos(18*u.deg), transform=proj_sun, linewidth=0, color='black', alpha=alpha)

    ptcolor = kwargs.get('ptcolor', ptcolor)
    lncolor = kwargs.get('lncolor', lncolor)
    ercolor = kwargs.get('ercolor', ercolor)
    rncolor = kwargs.get('rncolor', rncolor)
    atcolor = kwargs.get('atcolor', atcolor)
    outcolor = kwargs.get('outcolor', outcolor)
    hcolor = kwargs.get('hcolor', hcolor)

########### calculates path ##################
    vec = np.arange(0, int(8000/(np.absolute(occs['vel'].value))), step)
    vec = np.sort(np.concatenate((vec,-vec[1:]), axis=0))
    pa = Angle(occs['posa'])
    pa.wrap_at('180d', inplace=True)
    if pa > 90*u.deg:
        paplus = pa - 180*u.deg
    elif pa < -90*u.deg:
        paplus = pa + 180*u.deg
    else:
        paplus = pa
    deltatime = vec*u.s
    datas1 = data + deltatime
    centers_gcrs = GCRS(np.repeat(occs['stars'].ra, len(datas1)),
            np.repeat(occs['stars'].dec, len(datas1)), 1*u.R_earth, obstime=datas1)
    centers_itrs = centers_gcrs.transform_to(ITRS(obstime=datas1))
    centers = centers_itrs.earth_location

    dista = (occs['dist'].to(u.km)*ca1.to(u.rad)).value*u.km
    ax = dista*np.sin(pa) + (deltatime*occs['vel'])*np.cos(paplus)
    by = dista*np.cos(pa) - (deltatime*occs['vel'])*np.sin(paplus)

    ax2 = ax - (radius)*np.sin(paplus)
    by2 = by - (radius)*np.cos(paplus)
    lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
    j = np.where(lon1 < 1e+30)
    axf.plot(lon1[j], lat1[j], transform=ccrs.Geodetic(), color=lncolor)
    j = np.where(lon1 > 1e+30)
    if 'centerproj' not in kwargs:
        plt.plot(ax2[j].to(u.m).value, by2[j].to(u.m).value, color=outcolor, clip_on=(not centert), zorder=-0.2)

    ax3 = ax + (radius)*np.sin(paplus)
    by3 = by + (radius)*np.cos(paplus)
    lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
    j = np.where(lon2 < 1e+30)
    axf.plot(lon2[j], lat2[j], transform=ccrs.Geodetic(),  color=lncolor)
    j = np.where(lon2 > 1e+30)
    if 'centerproj' not in kwargs:
        plt.plot(ax3[j].to(u.m).value, by3[j].to(u.m).value, color=outcolor, clip_on=(not centert), zorder=-0.2)

##### plots error #####
    if erro is not None:
        err = erro*u.mas
        errd = (occs['dist'].to(u.km)*err.to(u.rad)).value*u.km
        ax2 = ax - errd*np.sin(paplus) - radius*np.sin(paplus)
        by2 = by - errd*np.cos(paplus) - radius*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(),  color=ercolor)

        ax3 = ax + errd*np.sin(paplus) + radius*np.sin(paplus)
        by3 = by + errd*np.cos(paplus) + radius*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(),  color=ercolor)

##### plots ring #####
    if ring is not None:
        rng = ring*u.km
        ax2 = ax - rng*np.sin(paplus)
        by2 = by - rng*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(),  color=rncolor)

        ax3 = ax + rng*np.sin(paplus)
        by3 = by + rng*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(),  color=rncolor)

##### plots atm #####
    if atm is not None:
        atmo = atm*u.km
        ax2 = ax - atmo*np.sin(paplus)
        by2 = by - atmo*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(),  color=atcolor)

        ax3 = ax + atmo*np.sin(paplus)
        by3 = by + atmo*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(),  color=atcolor)

##### plots center points #####
    vec = np.arange(0, int(8000/(np.absolute(occs['vel'].value))), cpoints)
    deltatime = np.sort(np.concatenate((vec,-vec[1:]), axis=0))*u.s
    axc = dista*np.sin(pa) + (deltatime*occs['vel'])*np.cos(paplus)
    byc = dista*np.cos(pa) - (deltatime*occs['vel'])*np.sin(paplus)
    plt.plot(axc.to(u.m).value, byc.to(u.m).value, 'o', color=ptcolor, clip_on=(not centert),
                 markersize=mapsize[0].value*pscale*8.0/46.0, zorder=-0.2)

    datas2 = data + deltatime
    centers_p_gcrs = GCRS(np.repeat(occs['stars'].ra, len(datas2)),
            np.repeat(occs['stars'].dec, len(datas2)), 1*u.R_earth, obstime=datas2)
    centers_p_itrs = centers_p_gcrs.transform_to(ITRS(obstime=datas2))
    centers_p = centers_p_itrs.earth_location
    clon1, clat1 = xy2latlon(axc.to(u.m).value, byc.to(u.m).value, centers_p.lon.value, centers_p.lat.value, datas2)
    j = np.where(clon1 < 1e+30)
    axf.plot(clon1[j], clat1[j],'o', transform=ccrs.Geodetic(), color=ptcolor, clip_on=True, markersize=mapsize[0].value*pscale*8.0/46.0)

    datas1 = data + deltatime
    center_gcrs = GCRS(np.repeat(occs['stars'].ra, 1),
            np.repeat(occs['stars'].dec, 1), 1*u.R_earth, obstime=data)
    center_itrs = center_gcrs.transform_to(ITRS(obstime=data))
    center = center_itrs.earth_location
    xp = [(dista.to(u.m)*np.sin(pa)).value]
    yp = [(dista.to(u.m)*np.cos(pa)).value]
    loncen, latcen = xy2latlon(xp, yp, center.lon.value, center.lat.value, data)
    j = np.where(loncen < 1e+30)
    if len(j) > 0:
        axf.plot(loncen[j], latcen[j],'o', transform=ccrs.Geodetic(), color=ptcolor, clip_on=True, markersize=mapsize[0].value*pscale*24.0/46.0)
    elif not centert:
        plt.plot(xp, yp, 'o', color=ptcolor, clip_on=False, markersize=mapsize[0].value*pscale*24.0/46.0)

######## plots the heights #########
    if 'heights' in kwargs.keys():
        for h in heights:
            plt.plot(bordx*np.cos(h*u.deg),bordy*np.cos(h*u.deg), linestyle='dotted', color=hcolor)

########### plots the the direction arrow ##################
    if arrow:
        if limits is None:
            plt.quiver(5500000,-5500000, (np.sin(paplus+90*u.deg)*np.sign(occs['vel'])).value,
                      (np.cos(paplus+90*u.deg)*np.sign(occs['vel'])).value, width=0.005)
        else:
            plt.quiver(lx + (ux-lx)*0.9,ly + (uy-ly)*0.1, (np.sin(paplus+90*u.deg)*np.sign(occs['vel'])).value,
                      (np.cos(paplus+90*u.deg)*np.sign(occs['vel'])).value, width=0.005, zorder = 1.3)

####### plots the countries names #####
    for country in countries.keys():
        plt.text(countries[country][0],countries[country][1], country, transform=ccrs.Geodetic(), weight='bold', color='grey', fontsize=30*cscale, family='monospace')

####### plots the sites ##############
    for site in sites.keys():
        s = EarthLocation.from_geodetic(sites[site][0]*u.deg,sites[site][1]*u.deg,0.0*u.km)
        axf.plot(s.lon.deg,s.lat.deg,'o', transform=ccrs.Geodetic(), markersize=mapsize[0].value*sscale*10.0/46.0, color=sites[site][4])
        xt,yt = latlon2xy(s.lon.deg, s.lat.deg, center_map.lon.value, center_map.lat.value)
        axf.text(xt + sites[site][2]*1000,yt+sites[site][3]*1000, site, weight='bold', fontsize=25*nscale, family='monospace')

######## Define the title and label of the output #########
    title = ('Object        Diam   Tmax   dots <> ra_offset_dec\n{:10s} {:4.0f} km  {:5.1f}s  {:02d} s <>{:+6.1f} {:+6.1f} \n'.
            format(name, 2*radius.value, (2*radius/np.absolute(occs['vel'])).value, cpoints, off_ra.value, off_de.value))
    labelx = ("\n year-m-d    h:m:s UT     ra__dec__J2000__candidate    C/A    P/A    vel   Delta   G*  long\n"
             "{}  {:02d} {:02d} {:07.4f} {:+03d} {:02d} {:06.3f} {:6.3f} {:6.2f} {:6.2f}  {:5.2f} {:5.1f}  {:3.0f}".
             format(data.iso,int(occs['stars'].ra.hms.h), int(occs['stars'].ra.hms.m), occs['stars'].ra.hms.s,
             int(occs['stars'].dec.dms.d), np.absolute(int(occs['stars'].dec.dms.m)), np.absolute(occs['stars'].dec.dms.s),
             ca1.value, occs['posa'].value, occs['vel'].value, occs['dist'].value, occs['magG'], occs['longi']))

####### plots the map#####
    if labels:
        axf.set_title(title, family='monospace', weight='bold', fontsize=22)
        axf.text(0.5, -0.1, labelx, va='bottom', ha='center', rotation='horizontal', rotation_mode='anchor', transform=axf.transAxes,
                 family='monospace', weight='bold', fontsize=22)
    plt.savefig('{}.{}'.format(nameimg, fmt), format=fmt, dpi=dpi)
    print('{}.{} generated'.format(nameimg, fmt))
    plt.clf()
    plt.close()




import numpy as np
import matplotlib.pylab as pl
import astropy.units as u
from astropy.timeseries import BoxLeastSquares
from astropy.time import Time
import scipy.special as scsp
import scipy.stats as scst
from datetime import datetime
from scipy.odr import odrpack as odr
from scipy.odr import models
from .extra import ChiSquare
import os
import warnings

def calc_fresnel(distance,lambida):
    """ Returns the fresnel scale.
    ----------
    Parameters
    ----------
    distance  (int, float, array): distances, in km.
    lambida   (int, float, array): wavelength, in km.
    ----------
    Return
    fresnel_scale  (float, array): fresnel scale, in km
    """
    return np.sqrt(lambida*distance/2)


def fit_pol(x,y,deg):
    """ Fit a polynom to the data.
    ----------
    Parameters
    ----------
    x  (array): 
    y  (array):
    deg  (int): order of the polynom 
    ----------
    Return
    ----------
    param (array): Array with the fitted values
    param_err (array): Array with the errors of the fitted values    
    """
    func   = models.polynomial(deg)
    mydata = odr.Data(x, y)
    myodr  = odr.ODR(mydata, func,maxit=200)
    myodr.set_job(fit_type=2)
    fit = myodr.run()
    param = fit.beta[::-1]                      
    param_err = fit.sd_beta[::-1]
    return param, param_err


class LightCurve():
    __names = []
    def __init__(self, name, **kwargs):
        """ Define a Light Curve

        Parameters:
            name (str): The name of the LightCurve. (required)
                Each time an LightCurve object is defined,
                the name must be different.
            tref (Time,str,float): Instant of reference. It can be
                in Julian Date, string in ISO format or Time object.
                Required only if LightCurve have input fluxes and
                given time is not in Julian Date.
            lambda (int,float): The band pass of observation (not required)
                in microns
            delta_lambda (int,float): The band width (not required.)

            Input data must be one of the 4 options below:
            file (str): a file with the time and flux in the first
                and second columns, respectively. A third columns
                with error in flux can also be given.

            IF file is not given:
            time: time must be a list of times, in seconds from tref,
                or Julian Date, or a Time object.
            flux: flux must be a list of fluxes. It must have the
                same lenght as time.
            dflux: if file not given, dflux must be a list of fluxes errors.
                It must have the same lenght as time. (not required)

            IF time and flux are not given.
            For a positive occultation
            immersion: The instant of immersion.
            emersion: The instant of emersion
            immersion_err: Immersion time uncertainty
            emersion_err: Emersion time uncertainty

            For a negative occultation
            initial_time: The initial time of observation
            end_time: The end time of observation.
        """
        input_done = False
        self.dflux = None
        self.__name = name
        self.flux = None
        self.time_model = None
        if self.__name in self.__names:
            raise ValueError('name {} already defined for another LightCurve object. Please choose a different one.'.format(self.__name))
        if 'immersion' in kwargs and 'emersion' in kwargs:
            if type(kwargs['immersion']) == str:
                self._immersion = Time(kwargs['immersion'])
            else:
                self._immersion = kwargs['immersion']
            if type(kwargs['emersion']) == str:
                self._emersion = Time(kwargs['emersion'])
            else:
                self._emersion = kwargs['emersion']
            self.immersion_err = 0.0
            if 'immersion_err' in kwargs:
                self.immersion_err = kwargs['immersion_err']
            self.emersion_err = 0.0
            if 'emersion_err' in kwargs:
                self.emersion_err = kwargs['emersion_err']
            input_done = True
        if 'initial_time' in kwargs and 'end_time' in kwargs:
            if type(kwargs['initial_time']) == str:
                self.initial_time = Time(kwargs['initial_time'])
            else:
                self.initial_time = kwargs['initial_time']
            if type(kwargs['end_time']) == str:
                self.end_time = Time(kwargs['end_time'])
            else:
                self.end_time = kwargs['end_time']
            input_done = True
        if not input_done:
            try:
                self.set_flux(**kwargs)
            except:
                raise ValueError('No allowed input conditions satisfied. Please refer to the tutorial.')
        self.lambda_0 = kwargs.get('lambda', 0.70)
        self.delta_lambda = kwargs.get('delta_lambda', 0.30)
        self.dt = 0.0
        self.__names.append(self.__name)

    @property
    def fresnel_scale(self):
        lamb  = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist  = self.dist*u.au.to('km')
        fresnel_scale_1 = calc_fresnel(dist,lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist,lamb+dlamb/2.0)
        fresnel_scale   = (fresnel_scale_1 + fresnel_scale_2)/2.0
        return fresnel_scale

    @property
    def name(self):
        return self.__name

    @property
    def immersion(self):
        return self._immersion + self.dt*u.s

    @property
    def emersion(self):
        return self._emersion + self.dt*u.s

    def check_names(self):
        return self.__names
    
    def set_flux(self, **kwargs):
        """ Set the flux for the LightCurve

        Parameters:
            exptime (int,float): The exposure time of the observation. (required)
            file (str): a file with the time and flux in the first
                and second columns, respectively. A third columns
                with error in flux can also be given.
            time: if file not given, time must be a list of times,
                in seconds from tref, or Julian Date, or a Time object.
            flux: if file not given, flux must be a list of fluxes.
                It must have the same lenght as time.
            dflux: if file not given, dflux must be a list of fluxes errors.
                It must have the same lenght as time.
            tref (Time,str,float): Instant of reference. It can be
                in Julian Date, string in ISO format or Time object.
        """
        input_done = False
        if 'file' in kwargs:
            if not os.path.isfile(kwargs['file']):
                raise ValueError('{} not found'.format(kwargs['file']))
            try:
                time, self.flux, self.dflux = np.loadtxt(kwargs['file'], usecols=[0,1,2], unpack=True)
                self.flux_obs = self.flux
            except:
                time, self.flux = np.loadtxt(kwargs['file'], usecols=[0,1], unpack=True)
                self.flux_obs = self.flux
            else:
                raise ValueError('Input file must have 2 or 3 columns')
            input_done = True
        if 'time' in kwargs and 'flux' in kwargs:
            if input_done:
                raise ValueError('Only one type of input can be given. Please refer to the tutorial.')
            self.flux = kwargs['flux']
            time = kwargs['time']
            if len(self.flux) != len(time):
                raise ValueError('time and flux must have the same length')
            if 'dflux' in kwargs:
                self.dflux = kwargs['dflux']
                if len(self.flux) != len(self.dflux):
                    raise ValueError('dflux must have the same length as flux and time')
            input_done = True
        if 'exptime' not in kwargs:
            raise ValueError('exptime not defined')
        self.exptime = kwargs['exptime']
        if 'tref' in kwargs:
            try:
                if type(kwargs['tref']) in [Time,str]:
                    self.tref = Time(kwargs['tref'])
                elif type(kwargs['tref']) in [int,float]:
                    self.tref = Time(kwargs['tref'], format='jd')
            except:
                raise ValueError('tref must be an Julian Date or ISO format Date')
        if 'time' in locals():
            if type(time) == Time:
                if not 'tref' in kwargs:
                    self.tref = Time(time[0].iso.split(' ')[0] + ' 00:00:00.000')
            elif all(time > 2400000):
                time = Time(time, format='jd')
                if not 'tref' in kwargs:
                    self.tref = Time(time[0].iso.split(' ')[0] + ' 00:00:00.000')
            elif not hasattr(self,'tref'):
                raise ValueError('tref must be given')
            else:
                time = self.tref + time*u.s
            self.time = (time - self.tref).sec
            order = np.argsort(self.time)
            self.model = np.ones(len(self.time))
            self.flux = self.flux[order]
            self.flux_obs = self.flux
            self.time = self.time[order]
            if self.dflux != None:
                self.dflux = self.dflux[order]
            self.initial_time = np.min(time)
            self.end_time = np.max(time)
            self.cycle = np.median(self.time[1:] - self.time[:-1])
            if self.cycle < self.exptime:
                warnings.warn('Exposure time ({:0.4f} seconds) higher than Cycle time ({:0.4f} seconds)'.format(self.exptime,self.cycle))

    def set_vel(self,vel):
        '''
        Set the occultation velocity
        Inputs:
        vel = float, in km/s
        '''
        if type(vel) == u.quantity.Quantity:
            vel = vel.to(u.km/u.s).value
        elif type(vel) in [float,int]:
            pass
        else:
            raise TypeError('vel must be an integer, a float or an Astropy Unit object')
        self.vel = np.absolute(vel)

    def set_dist(self,dist):
        '''
        Set the object distance
        Inputs:
        dist = float, in km
        '''
        if type(dist) == u.quantity.Quantity:
            dist = dist.to(u.AU).value
        elif type(dist) in [float,int]:
            pass
        else:
            raise TypeError('dist must be an integer, a float or an Astropy Unit object')
        self.dist = dist

    def set_star_diam(self,d_star):
        '''
        Set the star diameter
        Inputs:
        diam = float, in km
        '''
        if type(d_star) == u.quantity.Quantity:
            d_star = d_star.to(u.km).value
        elif type(d_star) in [float,int]:
            pass
        else:
            raise TypeError('d_star must be an integer, a float or an Astropy Unit object')
        self.d_star = d_star

    def set_filter(self,lambda_0,delta_lambda):
        '''
        Set the filter bandwidth in microns
        Inputs:
        lambda_0 = float, in microns
        delta_lambda = float, in microns
        '''
        if type(lambda_0) == u.quantity.Quantity:
            lambda_0 = lambda_0.to(u.micrometer).value
        elif type(lambda_0) in [float]:
            pass
        else:
            raise TypeError('lambda_0 must be a float or an Astropy Unit object')
        self.lambda_0 = lambda_0
        if type(delta_lambda) == u.quantity.Quantity:
            delta_lambda = delta_lambda.to(u.micrometer).value
        elif type(delta_lambda) in [float]:
            pass
        else:
            raise TypeError('delta_lambda must be a float or an Astropy Unit object')
        self.delta_lambda = delta_lambda

        
    def calc_magnitude_drop(self,mag_star,mag_obj):
        """
        Determine the magnitude drop
        ----------
        Parameters
        ----------
        mag_star (float): Stellar magnitude.
        mag_obj  (float): Object apparent magnitude to the date.
        ----------
        Returns
        ----------
        mag_drop (float): Magnitude drop for the given magnitudes     
        bottom_flux (float): Normalized bottom flux for the given magnitudes     
        """
        contrast= 1/(1+(10**((mag_star-mag_obj)*0.4)))
        mag_combined = mag_star-(2.5*(np.log10(1/contrast)))
        mag_drop = mag_obj - mag_combined
        bottom_flux = 10**((mag_combined - mag_obj)*0.4)
        self.mag_drop = mag_drop
        self.bottom_flux = bottom_flux
        return 
 
    def normalize(self,poly_deg=None,mask=None,**kwargs):
        """ Returns the fresnel scale.
        ----------
        Parameters
        ----------
        poly_deg  (int): degree of the polynom to be fitted
        mask   (array of Bolleans): which values to be fitted 
        ----------
        kwargs
        ----------
        flux_min (int,float): event flux to be setted as 0.0
        flux_max (int,float): baseline flux to be setted as 1.0        
        plot (Bollean): If True plot the steps for visual aid  
        """
        #Create a mask where the polynomial fit will be done
        if np.all(self.flux) == None:
            raise ValueError('Normalization is only possible when a LightCurve is instatiated with time and flux.')
        self.reset_flux()
        flux_min = 0.0
        flux_max = 1.0
        plot = False
        if 'flux_min' in kwargs:
            flux_min = kwargs['flux_min']
        if 'flux_max' in kwargs:
            flux_max = kwargs['flux_max']
        if 'plot' in kwargs:
            if (kwargs['plot'] == True):
                plot = True
        lc_flux = (self.flux - flux_min)/(flux_max-flux_min)
        if (np.all(mask) == None):
            preliminar_occ = self.occ_detect()
            tmax = preliminar_occ['emersion_time']+1.00*preliminar_occ['occultation_duration']
            tmin = preliminar_occ['immersion_time']-1.00*preliminar_occ['occultation_duration']
            chord = preliminar_occ['occultation_duration']
            mask = np.invert((self.time > tmin-(chord/2)) & (self.time < tmax+(chord/2)))
        norm_time = (self.time - self.time.min())/(self.time.max()-self.time.min())
        if (poly_deg != None):
            n = poly_deg
            p, err = fit_pol(norm_time[mask],lc_flux[mask],n)
            flux_poly_model = np.zeros(len(norm_time))
            for ii in np.arange(n+1):
                flux_poly_model = flux_poly_model + p[ii]*(norm_time**(n-ii))
            if (plot == True):
                pl.plot(norm_time[mask],lc_flux[mask],'k.-')
                pl.plot(norm_time[mask],flux_poly_model[mask],'r-')
                pl.title('Polynomial degree = {}'.format(n),fontsize=15)
                pl.show()       
        if (poly_deg == None):
            n = 0
            p, err = fit_pol(norm_time[mask],lc_flux[mask],n)
            flux_poly_model = np.zeros(len(norm_time))
            for ii in np.arange(n+1):
                flux_poly_model += p[ii]*(norm_time**(n-ii))
            if (plot == True):
                pl.plot(norm_time[mask],lc_flux[mask],'k.-')
                pl.plot(norm_time[mask],flux_poly_model[mask],'r-')
                pl.title('Polynomial degree = {}'.format(n),fontsize=15)
                pl.show()
            flux_poly_model_old = flux_poly_model.copy()
            for nn in np.arange(1,10):
                p, err = fit_pol(norm_time[mask],lc_flux[mask],nn)
                flux_poly_model_new = np.zeros(len(norm_time))
                for ii in np.arange(nn+1):
                    flux_poly_model_new += p[ii]*(norm_time**(nn-ii))
                F = np.var(flux_poly_model[mask]-lc_flux[mask])/np.var(flux_poly_model_new[mask]-lc_flux[mask])
                if (F > 1.05):
                    flux_poly_model = flux_poly_model_new.copy()
                    n = nn
                    if (plot == True):
                        pl.plot(norm_time[mask],lc_flux[mask],'k.-')
                        pl.plot(norm_time[mask],flux_poly_model[mask],'r-')
                        pl.title('Polynomial degree = {}'.format(nn),fontsize=15)
                        pl.show()
                else:
                    print('Normalization using a {} degree polynom'.format(n))
                    print('There is no improvement with a {} degree polynom'.format(n+1))
                    break
        self.flux = lc_flux/flux_poly_model
        self.normalizer_flux = flux_poly_model
        self.normalizer_mask = mask
        return

    def reset_flux(self):
        '''
        Reset flux for original values
        '''
        try:
            self.flux = self.flux_obs
        except:
            raise ValueError('Reset is only possible when a LightCurve is instatiated with time and flux.')
        return

    
    def occ_model(self,immersion_time, emersion_time, opacity, mask, npt_star=12, time_resolution_factor=10,flux_min=0,flux_max=1):
        """ Returns the modelled light curve considering fresnel difraction, star diameter and intrumental response.
        ----------
        Parameters
        ----------
        immersion_time (int, float): Immersion time, in seconds.                (input)
        emersion_time  (int, float): Emersion time, in seconds.                 (input)
        opacity (int, float): Opacity, opaque = 1.0, transparent = 0.0.         (input)
        mask (array with Booleans): Mask with True values to be computed        (input)
        npt_star  (int): Number of subdivisions for computing the star size's effects, default equal to 12. (auto)
        time_resolution_factor (int,float): Steps for fresnel scale used for modelling the light curve,     (auto)
            default equals to 10 steps for fresnel scale.
        flux_min (int,float): Bottom flux (only object), default equal to 0.0              (auto)
        flux_max (int,float): Base flux (object plus star), default equal to 1.0           (auto)
        ----------
        Returns
        ----------
        self.flux_inst       (array): Modelled Instrumental light flux.
        self.time_model      (array): Modelled timing.
        self.flux_star       (array): Modelled light flux considering fresnel difraction and star's diameter.
        self.flux_fresnel    (array): Modelled light flux considering fresnel difraction.
        self.model_geometric (array): Modelled light flux considering a box model.
        """
        # Computing the fresnel scale
        lamb  = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist  = self.dist*u.au.to('km')
        vel = np.absolute(self.vel)
        time_obs = self.time[mask]
        fresnel_scale_1 = calc_fresnel(dist,lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist,lamb+dlamb/2.0)
        fresnel_scale   = (fresnel_scale_1 + fresnel_scale_2)/2.0
        time_resolution = (np.min([fresnel_scale/vel,self.exptime]))/time_resolution_factor
        #
        #Creating a high resolution curve to compute fresnel difraction, stellar diameter and instrumental integration 
        time_model = np.arange(time_obs.min()-5*self.exptime,time_obs.max()+5*self.exptime,time_resolution)
        #
        #Changing X: time (s) to distances in the sky plane (km), considering the tangential velocity (vel in km/s)
        x   = time_model*vel    
        x01 = immersion_time*vel
        x02 = emersion_time*vel
        #
        #Computing fresnel diffraction for the case where the star size is negligenciable
        flux_fresnel_1 = self.__bar_fresnel(x,x01,x02,fresnel_scale_1,opacity)
        flux_fresnel_2 = self.__bar_fresnel(x,x01,x02,fresnel_scale_2,opacity)
        flux_fresnel   = (flux_fresnel_1 + flux_fresnel_2)/2.
        flux_star      = flux_fresnel.copy()
        if (self.d_star > 0):
            #Computing fresnel diffraction for the case where the star size is not negligenciable
            resolucao   = self.d_star/npt_star
            flux_star_1 = np.zeros(len(time_model))
            flux_star_2 = np.zeros(len(time_model))
            #Computing stellar diameter only near the immersion or emersion times
            star_diam = (np.absolute(x - x01) < 3*self.d_star) + (np.absolute(x - x02) < 3*self.d_star)
            p = np.arange(-npt_star,npt_star)*resolucao
            coeff = np.sqrt(np.absolute(self.d_star**2 - p**2))
            for ii in np.where(star_diam == True)[0]:
                xx = x[ii] + p
                flux1 = self.__bar_fresnel(xx,x01,x02,fresnel_scale_1,opacity)
                flux2 = self.__bar_fresnel(xx,x01,x02,fresnel_scale_2,opacity)
                flux_star_1[ii] = np.sum(coeff*flux1)/coeff.sum()
                flux_star_2[ii] = np.sum(coeff*flux2)/coeff.sum()
                flux_star[ii]   = (flux_star_1[ii] + flux_star_2[ii])/2.
        flux_inst = np.zeros(len(time_obs)) 
        for i in range(len(time_obs)):
            event_model  = (time_model > time_obs[i]-self.exptime/2.) & (time_model < time_obs[i]+self.exptime/2.)
            flux_inst[i] = (flux_star[event_model]).mean()
        self.model[mask] = flux_inst*(flux_max - flux_min) + flux_min
        self.time_model = time_model
        self.model_star = flux_star*(flux_max - flux_min) + flux_min
        self.model_fresnel = flux_fresnel*(flux_max - flux_min) + flux_min
        ev_model = (time_model > immersion_time) & (time_model < emersion_time)
        flux_box = np.ones(len(time_model))
        flux_box[ev_model] = (1-opacity)**2
        flux_box = flux_box*(flux_max - flux_min) + flux_min
        self.model_geometric = flux_box
        self.baseflux = flux_max
        self.bottomflux = flux_min
        return


  
    def occ_lcfit(self, **kwargs):
        """ Monte Carlo chi square fit for occultations lightcurve.
        ----------
        Parameters
        ----------
        tmin (int,float): Minimum time to consider in the fit prcedure                  
        tmax (int,float): Maximum time to consider in the fit prcedure                  
        flux_min (int,float): Bottom flux (only object), default equal to 0.0          
        flux_max (int,float): Base flux (object plus star), default equal to 1.0       
        immersion_time  (int, float): Initial guess for immersion time, in seconds.                       
        emersion_time   (int, float): Initial guess for emersion time, in seconds.                        
        opacity  (int, float): Initial guess for opacity, opaque = 1.0, transparent = 0.0, default equal to 0.0.   
        delta_t (int, float): Interval to fit immersion or emersion time               
        dopacity   (int, float): Interval to fit opacity, default equal to 0, no fit.  
        loop       (int): Number of tests to be done, default equal to 10000.          
        ----------
        Returns
        ----------
        chi2 (ChiSquare): ChiSquare object
        """
        #
        if not hasattr(self, 'flux'):
            raise ValueError('Fit curve is only possible when a LightCurve is instatiated with time and flux.')
        delta_t = 2*self.cycle
        loop = 10000
        t_i = np.zeros(loop)
        t_e = np.zeros(loop)
        tmax = self.time.max()
        tmin = self.time.min()
        immersion_time = tmin - self.exptime
        do_immersion = False
        emersion_time = tmax + self.exptime
        do_emersion = False
        opacity = 1.0
        delta_opacity = 0.0
        do_opacity = False
        if 'loop' in kwargs:
            loop = kwargs['loop']
        if ('immersion_time' not in kwargs) and ('emersion_time' not in kwargs):
            preliminar_occ = self.occ_detect()
            immersion_time = preliminar_occ['immersion_time']
            do_immersion = True
            emersion_time = preliminar_occ['emersion_time']
            do_emersion = True
            delta_t = 5*preliminar_occ['time_err']
            tmax = emersion_time+2*preliminar_occ['occultation_duration']
            tmin = immersion_time-2*preliminar_occ['occultation_duration']
            if 2*preliminar_occ['occultation_duration'] < 10*self.cycle:
                tmax = emersion_time + 10*self.cycle
                tmin = immersion_time - 10*self.cycle
        if 'tmax' in kwargs:
            tmax = kwargs['tmax']
        if 'tmin' in kwargs:
            tmin = kwargs['tmin']
        if 'delta_t' in kwargs:
            delta_t = kwargs['delta_t']
        if 'immersion_time' in kwargs:
            immersion_time = kwargs['immersion_time']
            do_immersion = True
        t_i = immersion_time + delta_t*(2*np.random.random(loop) - 1)
        if 'emersion_time' in kwargs:
            emersion_time = kwargs['emersion_time']
            do_emersion = True
        t_e = emersion_time  + delta_t*(2*np.random.random(loop) - 1)
        mask = (self.time >= tmin) & (self.time <= tmax)
        mask_sigma = (((self.time >= tmin) & (self.time < immersion_time - self.exptime)) +
                      ((self.time > emersion_time + self.exptime) & (self.time <= tmax)))
        sigma = self.flux[mask_sigma].std(ddof=1)
        if 'sigma' in kwargs:
            sigma = kwargs['sigma']
        if 'opacity' in kwargs:
            opacity = kwargs['opacity']
        if 'dopacity' in kwargs:
            delta_opacity = kwargs['dopacity']
            do_opacity = True
        opas = opacity + delta_opacity*(2*np.random.random(loop) - 1)
        opas[opas>1.], opas[opas<0.] = 1.0, 0.0
        flux_min = 0
        flux_max = 1
        if 'flux_min' in kwargs:
            flux_min = kwargs['flux_min']
        if 'flux_max' in kwargs:
            flux_max = kwargs['flux_max']
        #
        tflag = np.zeros(loop)
        tflag[t_i > t_e] = t_i[t_i > t_e]
        t_i[t_i > t_e]    = t_e[t_i > t_e]
        t_e[t_i > t_e]    = tflag[t_i > t_e]
        chi2 = 999999*np.ones(loop)
        #tcontrol_f0 = datetime.now()
        for i in range(loop):
            model_test = self.__occ_model(t_i[i],t_e[i],opas[i],mask,flux_min=flux_min,flux_max=flux_max)
            chi2[i] = np.sum((self.flux[mask] -  model_test)**2)/(sigma**2)
        #tcontrol_f3 = datetime.now()
        #print('Elapsed time: {:.3f} seconds.'.format((tcontrol_f3 - tcontrol_f0).total_seconds()))
        kkargs = {}
        if do_immersion:
            kkargs['immersion'] = t_i
        if do_emersion:
            kkargs['emersion'] = t_e
        if do_opacity:
            kkargs['opacity'] = opas
        chisquare = ChiSquare(chi2, len(self.flux[mask]), **kkargs)
        onesigma = chisquare.get_nsigma(1)
        if 'immersion' in onesigma:
            self._immersion = self.tref + onesigma['immersion'][0]*u.s
            self.immersion_err = onesigma['immersion'][1]
            immersion_time = onesigma['immersion'][0]
        if 'emersion' in onesigma:
            self._emersion = self.tref + onesigma['emersion'][0]*u.s
            self.emersion_err = onesigma['emersion'][1]
            emersion_time = onesigma['emersion'][0]
        if 'opacity' in onesigma:
            opacity = onesigma['opacity'][0]
        # Run occ_model() to save best parameters in the Object.
        self.occ_model(immersion_time,emersion_time,opacity,mask,flux_min=flux_min,flux_max=flux_max)
        self.lc_sigma = sigma
        self.chisquare = chisquare
        return chisquare

    def plot_lc(self):
        '''
        Plot the light curve
        '''
        if (np.any(self.flux) != None):
            pl.close()
            pl.plot(self.time,self.flux,'k.-',label='Obs.',zorder=0)
            if (np.any(self.model) != None):
                pl.plot(self.time,self.model,'r-',label='Model',zorder=2)
                pl.scatter(self.time,self.model, s=50, facecolors='none', edgecolors='r',zorder=3)
            pl.tight_layout()
            pl.xlabel('Time [seconds]',fontsize=20)
            pl.ylabel('Relative Flux',fontsize=20)
            pl.legend()
        else:
            raise ValueError('Plotting the light curve is only possible when the Object LightCurve is instatiated with time and flux')
        return

    def plot_model(self):
        '''
        Plot the modelled light curve
        '''
        if np.all(self.time_model) != None:
            pl.plot(self.time_model,self.model_geometric,'c-',label='Geometric',zorder=1)
            pl.plot(self.time_model,self.model_fresnel,'b-',label='Fresnel',zorder=1)
            pl.plot(self.time_model,self.model_star,'g-',label='Star diam.',zorder=1)
            pl.tight_layout()
            pl.xlabel('Time [seconds]',fontsize=20)
            pl.ylabel('Relative Flux',fontsize=20)
            pl.legend()
        else:
            raise ValueError('Plotting the model light curve is only possible after the model [LightCurve.occ_model()] or the fit [LightCurve.occ_lcfit()]')
        return

    def to_log(self,namefile=None):
        """ Save the light curve log to a file

        Parameters:
            namefile (str): Filename to save the log
        """
        if (namefile == None):
            namefile = self.name.replace(' ','_')+'.log'
        f = open(namefile, 'w')
        f.write(self.__str__())
        f.close()
        

    def to_file(self, namefile=None):
        """ Save the light curve to a file

        Parameters:
            namefile (str): Filename to save the data
        """
        ###### Observational data
        if (namefile == None):
            folder=''
            file = self.name.replace(' ','_')+'.dat'
        else:
            folder = os.path.dirname(namefile)
            file = os.path.basename(namefile)
        data = np.array([(self.time*u.s + self.tref).jd,self.time,self.flux,self.model,self.flux-self.model])
        colunm_names = ['Time JD','Time relative to {} UTC in seconds'.format(self.tref.iso),'Observational Flux','Modelled Flux','Residual O-C']
        np.savetxt(os.path.join(folder,file), data.T, fmt='%11.8f')
        f = open(os.path.join(folder,file) + '.label', 'w')
        for i, name in enumerate(colunm_names):
            f.write('Column {}: {}\n'.format(i+1,name))
        f.close()
        ###### Complete Model
        if (np.all(self.time_model) != None):
            data_model = np.array([(self.time_model*u.s + self.tref).jd,self.time_model,self.model_geometric,self.model_fresnel,self.model_star])
            colunm_names_model = ['Model time JD','Model time relative to {} UTC in seconds'.format(self.tref.iso),'Geometric Model','Model with Fresnel diffraction','Model with star diameter']
            np.savetxt(os.path.join(folder,'model_'+file), data_model.T, fmt='%11.8f')
            f = open(os.path.join(folder,'model_'+file)+'.label', 'w')
            for i, name in enumerate(colunm_names_model):
                f.write('Column {}: {}\n'.format(i+1,name))
            f.close()



    
    def occ_detect(self, maximum_duration=None, dur_step=None, snr_limit=None, \
                  n_detections=None, plot=False):
        """
        Detect automatically the occultation event in the light curve
        
        Parameters
        ----------
        maximum_duration (float): Maximum duration of the occultation event (default is the light curve's time span). (optional)
        dur_step (float): Step size to sweep occultation duration event (default value is 1/2 of sampling). (optional)
        snr_limit (float): Minimum occultation SNR. (optional)
        n_detections (int): N best detections regardless from SNR. n_detections is superseded by snr_limit. (optional)
        plot (boolean): True if output plots are desired. (optional)
            
        Returns
        -------
        OrderedDict
            An ordered dictionary of :attr:`name`::attr:`value` pairs for each Parameter.
            
        Examples
        --------
        >>> lc = sora.LightCurve(time=time, flux=flux, exptime=0.0, name='lc_example')
        >>> params = lc.occ_detect()
        >>> params
        {'rank': 1, 
        'occultation_duration': 40.1384063065052, 
        'central_time': 7916.773870512843, 
        'immersion_time': 7896.7046673595905, 
        'emersion_time': 7936.843073666096, 
        'time_err': 0.05011036992073059, 
        'depth': 0.8663887801707082, 
        'depth_err': 0.10986223384336465, 
        'baseline': 0.9110181732552853, 
        'baseline_err': 0.19045768512595365, 
        'snr': 7.886138392251848,
        'occ_mask': array([False, False, False, ..., False, False, False])}
        """

        if not hasattr(self, 'flux'):
            raise ValueError('time and flux must be instantiated to use '\
                             'occ_detect function.')

        # duration of the light curve
        time_span = self.time[-1]-self.time[0]
        if maximum_duration and (maximum_duration > time_span):
            warnings.warn('Occultation duration (maximum_duration={0}) ' \
                             'exceeds the time series lenght ({1:0.5f}).' \
                          ' maximum_duration reset to the time series lenght.'\
                          .format(maximum_duration, time_span))
            maximum_duration = time_span
        if not maximum_duration:
            maximum_duration = time_span

        if not dur_step:
            dur_step = self.cycle/2
        
        if dur_step < self.cycle/2:
            warnings.warn('The given dur_step is oversampled by a factor ' \
                          'of {0:0.1f} and has been reset to half a cycle ' \
                          '({1:0.4f}).' \
                          .format((self.cycle/2.)/dur_step, self.cycle/2.))
            dur_step = self.cycle/2
       
        duration_grid = np.arange(dur_step, maximum_duration, dur_step)
        # initial occultation mask (all data points)
        mask = np.ones(len(self.time), dtype=bool)
        # inital detection rank
        rank = 1

        if snr_limit:
            # minimum SNR accepted in a detection for multiple search
            snr_value = snr_limit+1
            occ0 = self.__run_bls(time_span, duration_grid)
            mask *= ~occ0['occ_mask']
            while (snr_value > snr_limit):
                rank += 1
                occ1 = self.__run_bls(time_span, duration_grid, mask=mask, \
                                      rank=rank)
                if occ1['snr'] > snr_limit:
                    snr_value = occ1['snr']
                    mask *= ~occ1['occ_mask']                    
                    occ0 = self.__summarize_bls(occ0,occ1)
                else:
                    snr_value = snr_limit
            
            if plot:
                self.__plot_occ_detect(occ0)                    
            return occ0
        elif n_detections:
            # search the n best fits
            occ0 = self.__run_bls(time_span, duration_grid)
            mask *= ~occ0['occ_mask']    
            for i in range(n_detections-1):
                rank += 1
                occ1 = self.__run_bls(time_span, duration_grid, mask=mask, \
                                      rank=rank)
                snr_value = occ1['snr']
                mask *= ~occ1['occ_mask']                    
                occ0 = self.__summarize_bls(occ0,occ1)
            
            if plot:
                self.__plot_occ_detect(occ0)
            return occ0
        else:
            # search only the first best fit
            occ0 = self.__run_bls(time_span, duration_grid)
           
            if plot:
                self.__plot_occ_detect(occ0)
                
            return occ0
        
    def __plot_occ_detect(self, occ):
        n = np.size(occ['rank'])
        if n > 1:
            # case for multiple detections
            pl.plot(self.time, self.flux, 'k.-')
            mask = np.zeros(len(self.time), dtype=bool)
            for i in range(n):
                trues = np.sum(occ['occ_mask'][i])
                pl.plot(self.time[occ['occ_mask'][i]], np.repeat(np.mean(self.flux[occ['occ_mask'][i]]),trues), '.', label='Rank: '+str(i+1))
                mask += occ['occ_mask'][i]
 
            falses = list(mask).count(False)
            pl.plot(self.time[~mask], np.repeat(np.mean(self.flux[~mask]),falses), 'r.', label='Baseline')
            pl.xlabel('Time [seconds]')
            pl.ylabel('Relative Flux')
            pl.legend()    
            
        else:
            # case for single occultation
            trues = list(occ['occ_mask']).count(True)
            falses = list(occ['occ_mask']).count(False)
            pl.plot(self.time, self.flux, 'k.-')
            pl.plot(self.time[occ['occ_mask']], np.repeat(np.mean(self.flux[occ['occ_mask']]),trues), '.', label='Occultation')
            pl.plot(self.time[~occ['occ_mask']], np.repeat(np.mean(self.flux[~occ['occ_mask']]),falses), 'r.', label='Baseline')
            pl.xlabel('Time [seconds]')
            pl.ylabel('Relative Flux')
            pl.legend()
            
    
    def __run_bls(self, per_grid, dur_grid, mask=None, rank=None):
        """
        Private function to find the best box fit suitable to the data
        """

        # object with no occultation mask
        mskmodel = BoxLeastSquares(self.time, self.flux, dy=self.dflux)
        # if there is no dflux array, reset it to None in case of 
        # using a mask
        if self.dflux is None:
            dfluxmask = None
        else:
            dfluxmask = self.dflux[mask]
            
        # object with occultation mask
        if np.sum(mask):
            model = BoxLeastSquares(self.time[mask], self.flux[mask], \
                                    dy=dfluxmask)
        else:
            model = mskmodel   
        
        r = model.power(per_grid, dur_grid, objective='snr', method='fast')
        # statistics of the BLS fit
        stats = model.compute_stats(r.period, r.duration, r.transit_time)
        # occultation mask of the event with respect to all data
        occ_mask = mskmodel.transit_mask(self.time, r.period, r.duration, \
                                         r.transit_time)
        # parameters computation for clarity purposes
        occultation_duration = r.duration[0]
        central_time = stats['transit_times'][0]
        immersion_time = stats['transit_times'][0] - r.duration[0]/2
        emersion_time = stats['transit_times'][0] + r.duration[0]/2
        time_err = np.median(self.time[1:-1]-self.time[0:-2])/2
        depth = np.mean(self.flux[~occ_mask])-np.mean(self.flux[occ_mask])
        depth_err = np.std(self.flux[occ_mask],ddof=1)
        baseline = np.mean(self.flux[~occ_mask])
        baseline_err = np.std(self.flux[~occ_mask],ddof=1)
        # If there is only one measurement during the occultation it will
        # use the baseline_err to compute SNR, otherwise it will use depth_err
        if np.sum(occ_mask) < 2:
            snr = depth/baseline_err
        else:
            snr = depth/depth_err


        # define rank
        if rank:
            rank = rank
        else:
            rank = 1
        
        return {'rank' : rank,
                'occultation_duration' : occultation_duration, 
                'central_time' : central_time, 
                'immersion_time' : immersion_time, 
                'emersion_time' : emersion_time,
                'time_err' : time_err,
                'depth' : depth, 'depth_err' : depth_err,
                'baseline' : baseline, 'baseline_err' : baseline_err,
                'snr' : snr,'occ_mask' : occ_mask}
    
    def __summarize_bls(self, dict1, dict2):
        ''' Private function to merge dictionaries returned by BLS and 
            to keep values of common keys in list.
        '''
        dict3 = {}
        for key, value in dict1.items():
            if key == 'occ_mask':
                sz = int(np.size(dict1[key])/np.size(dict2[key]))
                if sz > 1:
                    l = [None]*(sz+1)
                    for i in range(sz):
                        l[i] = dict1[key][i]
                    l[i+1] = dict2[key]
                    dict3[key] = l
                else:
                    dict3[key] = [dict1[key],dict2[key]]
            else:
                dict3[key] = np.append(dict1[key],dict2[key])
        return dict3

    def __bar_fresnel(self,X,X01,X02,fresnel_scale,opacity):
        """ Returns the modelled light curve considering fresnel difraction.
        ----------
        Parameters
        ----------
        X   (array): Array with time values converted in km using the event velocity.
        X01 (int, float): Immersion time converted in km using the event velocity.
        X02 (int, float): Emersion time converted in km using the event velocity.
        fresnel_scale (int, float): Fresnel scale.
        opacity (int, float): Opacity, opaque = 1.0, transparent = 0.0
        ----------
        Returns
        ----------
        flux_fresnel (array):
        """
        # Converting from km to units of fresnel scale
        x   = X/fresnel_scale
        x01 = X01/fresnel_scale
        x02 = X02/fresnel_scale
        # Fresnel difraction parameters 
        x1 = x - x01
        x2 = x - x02
        s1,c1 = scsp.fresnel(x1)
        s2,c2 = scsp.fresnel(x2)
        cc = c1 - c2
        ss = s1 - s2
        r_ampli = - (cc+ss)*(opacity/2.)
        i_ampli =   (cc-ss)*(opacity/2.)
        # Determining the flux considering fresnel difraction
        flux_fresnel = (1.0 + r_ampli)**2 + (i_ampli)**2
        return flux_fresnel


    def __occ_model(self,immersion_time, emersion_time, opacity, mask, npt_star=12, time_resolution_factor=10,flux_min=0.0,flux_max=1.0):
        """ Private function returns the modelled light curve considering fresnel difraction, star diameter and intrumental response, intended for fitting inside the self.occ_lcfit().
        ----------
        Parameters
        ----------
        immersion_time (int, float): Immersion time, in seconds.                (input)
        emersion_time  (int, float): Emersion time, in seconds.                 (input)
        opacity (int, float): Opacity, opaque = 1.0, transparent = 0.0.       (input)
        mask (array with Booleans): Mask with True values to be computed        (input)
        npt_star  (int): Number of subdivisions for computing the star size's effects, default equal to 12. (auto)
        time_resolution_factor (int,float): Steps for fresnel scale used for modelling the light curve,     (auto)
            default equals to 10 steps for fresnel scale.
        flux_min (int,float): Bottom flux (only object), default equal to 0.0              (auto)
        flux_max (int,float): Base flux (object plus star), default equal to 1.0           (auto)
        ----------
        Returns
        ----------
        flux_inst       (array): Modelled Instrumental light flux.
        """
        # Computing the fresnel scale
        lamb  = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist  = self.dist*u.au.to('km')
        vel = np.absolute(self.vel)
        time_obs = self.time[mask]
        fresnel_scale_1 = calc_fresnel(dist,lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist,lamb+dlamb/2.0)
        fresnel_scale   = (fresnel_scale_1 + fresnel_scale_2)/2.0
        time_resolution = (np.min([fresnel_scale/vel,self.exptime]))/time_resolution_factor
        self.model_resolution = time_resolution
        #
        #Creating a high resolution curve to compute fresnel difraction, stellar diameter and instrumental integration 
        time_model = np.arange(time_obs.min()-5*self.exptime,time_obs.max()+5*self.exptime,time_resolution)
        #
        #Changing X: time (s) to distances in the sky plane (km), considering the tangential velocity (vel in km/s)
        x   = time_model*vel    
        x01 = immersion_time*vel
        x02 = emersion_time*vel
        #
        #Computing fresnel diffraction for the case where the star size is negligenciable
        flux_fresnel_1 = self.__bar_fresnel(x,x01,x02,fresnel_scale_1,opacity)
        flux_fresnel_2 = self.__bar_fresnel(x,x01,x02,fresnel_scale_2,opacity)
        flux_fresnel   = (flux_fresnel_1 + flux_fresnel_2)/2.
        flux_star      = flux_fresnel.copy()
        if (self.d_star > 0):
            #Computing fresnel diffraction for the case where the star size is not negligenciable
            resolucao   = self.d_star/npt_star
            flux_star_1 = np.zeros(len(time_model))
            flux_star_2 = np.zeros(len(time_model))
            #Computing stellar diameter only near the immersion or emersion times
            star_diam = (np.absolute(x - x01) < 3*self.d_star) + (np.absolute(x - x02) < 3*self.d_star)
            p = np.arange(-npt_star,npt_star)*resolucao
            coeff = np.sqrt(np.absolute(self.d_star**2 - p**2))
            for ii in np.where(star_diam == True)[0]:
                xx = x[ii] + p
                flux1 = self.__bar_fresnel(xx,x01,x02,fresnel_scale_1,opacity)
                flux2 = self.__bar_fresnel(xx,x01,x02,fresnel_scale_2,opacity)
                flux_star_1[ii] = np.sum(coeff*flux1)/coeff.sum()
                flux_star_2[ii] = np.sum(coeff*flux2)/coeff.sum()
                flux_star[ii]   = (flux_star_1[ii] + flux_star_2[ii])/2.
        flux_inst = np.zeros(len(time_obs)) 
        for i in range(len(time_obs)):
            event_model  = (time_model > time_obs[i]-self.exptime/2.) & (time_model < time_obs[i]+self.exptime/2.)
            flux_inst[i] = (flux_star[event_model]).mean()
        return flux_inst*(flux_max - flux_min) + flux_min

    def __str__(self):
        """ String representation of the LightCurve Object
        """
        output  = 'Light curve name: {}\n'.format(self.name)
        output += 'Initial time: {} UTC\n'.format(self.initial_time.iso)
        output += 'End time:     {} UTC\n'.format(self.end_time.iso)
        output += 'Duration:     {:.3f} minutes\n'.format((self.end_time - self.initial_time).value*u.d.to('min'))
        output += 'Time offset:  {:.3f} seconds\n\n'.format(self.dt)
        try:
            output += 'Exposure time:    {:.4f} seconds\n'.format(self.exptime)
            output += 'Cycle time:       {:.4f} seconds\n'.format(self.cycle)
            output += 'Num. data points: {}\n\n'.format(len(self.time))
        except:
            output += 'Object LightCurve was not instantiated with time and flux.\n\n'
        try:
            output += 'Bandwidth:            {:.3f} +/- {:.3f} microns\n'.format(self.lambda_0,self.delta_lambda)
            output += 'Used shadow velocity: {:.3f} km/s\n'.format(self.vel)
            output += 'Fresnel scale:        {:.3f} seconds or {:.2f} km\n'.format(self.fresnel_scale/self.vel,self.fresnel_scale)
            output += 'Stellar size effect:  {:.3f} seconds or {:.2f} km\n'.format(self.d_star/self.vel,self.d_star)
        except:
            output += '\nThere is no occultation associated with this light curve.\n'
        try:
            output += 'Inst. response:       {:.3f} seconds or {:.2f} km\n'.format(self.exptime, self.exptime*self.vel)
            output += 'Dead time effect:     {:.3f} seconds or {:.2f} km\n'.format(self.cycle-self.exptime,(self.cycle-self.exptime)*self.vel)
            output += 'Model resolution:     {:.3f} seconds or {:.2f} km\n'.format(self.model_resolution, self.model_resolution*self.vel)
            output += 'Modelled baseflux:    {:.3f}\n'.format(self.baseflux)
            output += 'Modelled bottomflux:  {:.3f}\n'.format(self.bottomflux)
            output += 'Light curve sigma:    {:.3f}\n\n'.format(self.lc_sigma)
        except:
            output += '\nObject LightCurve model was not fitted.\n\n'
        try:
            output += 'Immersion time: {} UTC +/- {:.3f} seconds\n'.format(self.immersion.iso,self.immersion_err)
            output += 'Emersion time:  {} UTC +/- {:.3f} seconds\n\n'.format(self.emersion.iso,self.emersion_err)
        except:
            output += 'Immersion and emersion times were not fitted or instantiated.\n\n'

        try:
            output += 'Monte Carlo chi square fit.\n\n' + self.chisquare.__str__() + '\n'
        except:
            pass
        return output


    def __del__(self):
        """ When this object is deleted, it removes the name from the Class name list.
        """
        try:
            self.__names.remove(self.__name)
        except:
            pass
